<?php

/********************************************
* LiteSpeed AP2ZConf project
* @Author:   LiteSpeed Technologies, Inc. (http://www.litespeedtech.com)
* @Copyright: (c) 2017
*********************************************/

require_once('VHostConf.php');

class LSWSConf
{
	const XML_NAME = 'TAG';
	const XML_ELEM = 'ELEMENTS';
	const XML_VAL = 'VALUE';

	/**
	 * Load and parse the LSWS configurations.
	 *
	 * @param string $serverRoot Absolute path to the LSWS server root.
	 * @return array Returns compiled LSWS configuration.
	 */
	function loadServConf($serverRoot)
	{
		$file = $serverRoot . "/conf/httpd_config.xml";

		$xmltree = new XmlTreeBuilder();
		$rootNode = &$xmltree->parseFile($file);

		if ($rootNode == NULL)
			return array();

		return $this->extractServConf($rootNode);
	}

	/**
	 * Extract the LSWS server conf into an associated array.
	 *
	 * @param array $rootNode Associated array of configurations extracted from XML.
	 * @return array Returns compiled LSWS configuration.
	 */
	function &extractServConf(&$rootNode)
	{
		$serv = array(
			'general' => array(),
		);
		$elements = &$rootNode[self::XML_ELEM];
		for ($i = 0; $i < count($elements); ++$i)
		{
			$name = $elements[$i][self::XML_NAME];
			$children = &$elements[$i][self::XML_ELEM];

			switch ($name)
			{
			case 'serverName':   case 'adminEmails': case 'mime':
			case 'autoRestart':  case 'chrootPath':  case 'enableChroot':
			case 'inMemBufSize': case 'swappingDir': case 'user':
			case 'group':        case 'adminRoot':   case 'indexFiles':
			case 'showVersionNumber': case 'priority':
				$serv['general'][$name] = $elements[$i][self::XML_VAL];
				break;
			case 'logging':
				$this->extractLogging($children, $serv['general']);
				break;
			case 'htAccess':
				$this->extractHtAccess($children, $serv['general']);
				break;
			case 'expires':
				$this->extractExpires($children, $serv['general']);
				break;
			case 'tuning':
				$serv['tuning'] = array_column($children, self::XML_VAL, self::XML_NAME);
				break;
			case 'security':
				$serv['security'] = &$this->extractServSecurity($children);
				break;
			case 'virtualHostList':
				$this->extractVHosts($children, $serv);
				break;
			case 'listenerList':
				$this->extractListeners($children, $serv);
				break;
			case 'extProcessorList':
				$serv['ext'] = &$this->extractExt($elements[$i][self::XML_ELEM]);
				break;
			case 'scriptHandlerList':
				$serv['scriptHandler'] = &$this->extractScriptHandler($elements[$i][self::XML_ELEM]);
				break;
			}
		}
		return $serv;
	}

	/**
	 * Extract the Script Handlers configurations from the XML config.
	 *
	 * @param array $handlerList Array of Script Handlers configurations.
	 * @return array The compiled array of the extracted configurations.
	 */
	function &extractScriptHandler(&$handlerList)
	{
		$scriptHandler = array();
		$tags = array('suffix', 'type', 'handler');
		for ($i = 0; $i < count($handlerList); ++$i)
		{
			$handler = &$handlerList[$i][self::XML_ELEM];
			$sc = array();
			for ($j = 0; $j < count($handler); ++$j)
			{
				$name = $handler[$j][self::XML_NAME];
				if (in_array($name, $tags))
					$sc[$name] = $handler[$j][self::XML_VAL];
			}
			if (isset($sc['type']) && isset($sc['suffix']))
			{
				$scriptHandler[$sc['suffix']] = $sc;
			}
		}
		return $scriptHandler;
	}

	/**
	 * Extract the External Processors List configurations from the XML config.
	 *
	 * @param array $processorList Array of External Processors configurations.
	 * @return array The compiled array of the extracted configurations.
	 */
	function &extractExt(&$processorList)
	{
		$extProcList = array();
		$tags = array('address', 'name', 'type', 'maxConns', 'initTimeout',
			'retryTimeout', 'respBuffer', 'autoStart',
			'backlog', 'path', 'instances', 'priority');
		for ($i = 0; $i < count($processorList); ++$i)
		{
			$extProcEntry = array();
			$processor = &$processorList[$i][self::XML_ELEM];
			for ($j = 0; $j < count($processor); ++$j)
			{
				$name = $processor[$j][self::XML_NAME];
				if (in_array($name, $tags))
				{
					$extProcEntry[$name] = $processor[$j][self::XML_VAL];
				}
				elseif ($name == 'env')
				{
					if (!isset($extProcEntry['env'])) {
						$extProcEntry['env'] = array();
					}
					$extProcEntry['env'][] = $processor[$j][self::XML_VAL];
				}
			}
			if (isset($extProcEntry['name']))
				$extProcList[$extProcEntry['name']] = &$extProcEntry;
			unset($extProcEntry);
		}
		return $extProcList;
	}

	/**
	 * Extract the Server Security configurations from the XML config.
	 *
	 * @param array $securityConf Array of Server Security configurations.
	 * @return array The compiled array of the extracted configurations.
	 */
	function &extractServSecurity(&$securityConf)
	{
		$sec = array();
		for ($i = 0; $i < count($securityConf); ++$i)
		{
			$name = $securityConf[$i][self::XML_NAME];
			$tmpConf = array_column($securityConf[$i][self::XML_ELEM],
					self::XML_VAL, self::XML_NAME);
			switch ($name)
			{
			case 'fileAccessControl':
				$tags = array('followSymbolLink', 'checkSymbolLink',
					'requiredPermissionMask', 'restrictedPermissionMask');
				break;
			case 'CGIRLimit':
				$tags = array('maxCGIInstances', 'minUID', 'minGID', 'priority',
					'CPUSoftLimit', 'CPUHardLimit', 'memSoftLimit',
					'memHardLimit', 'procSoftLimit', 'procHardLimit');
				break;
			case 'perClientConnLimit':
				$tags = array('staticReqPerSec', 'dynReqPerSec',
								'outBandwidth', 'inBandwidth', 'softLimit',
								'hardLimit', 'gracePeriod', 'banPeriod');
				if (isset($tmpConf['throttleLimit']))
				{
					$sec['perClientConnLimit_outBandwidth'] = $tmpConf['throttleLimit'];
					$sec['perClientConnLimit_inBandwidth'] = $tmpConf['throttleLimit'];
				}
				break;
			case 'accessDenyDir':
				$denyDirs = &$securityConf[$i]['ELEMENTS'];
				$dirCnt = count($denyDirs);
				for ($j = 0; $j < $dirCnt; ++$j)
				{
					if ($denyDirs[$j][self::XML_NAME] == 'dir')
						$sec[$name][] = $denyDirs[$j][self::XML_VAL];
				}
				continue 2;
			case 'accessControl':
				if (isset($tmpConf['allow']))
					$sec['accessControl_allow'] = $tmpConf['allow'];
				if (isset($tmpConf['deny']))
					$sec['accessControl_deny'] = $tmpConf['deny'];
				continue 2;
			default:
				continue 2;
			}

			foreach ($tags as $tag)
			{
				if (isset($tmpConf[$tag]))
					$sec[$name . '_' . $tag] = $tmpConf[$tag];
			}
		}
		return $sec;
	}

	/**
	 * Extract the htaccess configurations from the XML config.
	 *
	 * @param array $el Array of htaccess configurations.
	 * @param array $holder The destination array to put the extracted configurations.
	 */
	function extractHtAccess(&$el, &$holder)
	{
		$tags = array('allowOverride', 'accessFileName');
		$this->extractData($el, $holder, $tags);
	}

	/**
	 * Extract the self::XML_ELEM array and inserts them into holder with the given prefix.
	 * 
	 * Only entries in the $tags array are extracted.
	 *
	 * @param array $elems The elements in the configuration.
	 * @param array $holder The destination array to put the extracted configurations.
	 * @param array $tags The desired tags from the elements array.
	 * @param string $entryPrefix The prefix to assign to the entries when inserted
	 * into holder.
	 */
	function extractData(&$elems, &$holder, &$tags, $entryPrefix = '')
	{
		$tmpConf = array_column($elems, self::XML_VAL, self::XML_NAME);

		foreach ($tags as $tag)
		{
			if (isset($tmpConf[$tag]))
				$holder[$entryPrefix . $tag] = $tmpConf[$tag];
		}
	}

	/**
	 * Extract the Expires configurations from the XML config.
	 *
	 * @param array $el Array of expires configurations.
	 * @param array $holder The destination array to put the extracted configurations.
	 */
	function extractExpires(&$el, &$holder)
	{
		$tags = array('enableExpires', 'expiresByType', 'expiresDefault');
		$this->extractData($el, $holder, $tags);
	}

	/**
	 * Extract the log configurations.
	 *
	 * @param array $logConf Array of log configurations.
	 * @param array $holder The destination array to put the extracted configurations.
	 */
	function extractLogging(&$logConf, &$holder)
	{
		$logTags = array('fileName', 'useServer', 'logLevel',
			'debugLevel', 'rollingSize', 'enableStderrLog');
		$aclogTags = array('fileName', 'useServer', 'keepDays',
			'rollingSize', 'logReferer', 'logUserAgent',
			'compressArchive');
		for ($i = 0; $i < count($logConf); ++$i)
		{
			$name = $logConf[$i][self::XML_NAME];
			if ($name == 'log')
			{
				$this->extractData($logConf[$i][self::XML_ELEM], $holder,
						$logTags, 'log_');
			}
			elseif ($name == 'accessLog')
			{
				$this->extractData($logConf[$i][self::XML_ELEM], $holder,
						$aclogTags, 'accessLog_');
			}
		}
	}

	/**
	 * Extract the listener configurations from the XML config.
	 *
	 * The configs are processed and a VHostConf is created for each listener.
	 *
	 * @param array $listenerList Array of listener configurations.
	 * @param array $holder The destination array to put the extracted configurations.
	 */
	function extractListeners(&$listenerList, &$holder)
	{
		for ($i = 0; $i < count($listenerList); ++$i)
		{
			$listenerArray = array_column($listenerList[$i][self::XML_ELEM],
					self::XML_VAL, self::XML_NAME);

			if (!isset($listenerArray['name']))
				continue;

			if (isset($listenerArray['address']))
			{
				$addr = $listenerArray['address'];
				$pos = strpos($addr, ':');
				if ($pos)
				{
					$listenerArray['ip'] = substr($addr, 0, $pos);
					$listenerArray['port'] = substr($addr, $pos + 1);
					if ($listenerArray['ip'] == '*')
						$listenerArray['ip'] = 'ANY';
				}
				unset($listenerArray['address']);
			}
			$listener = new VHostConf($listenerArray['ip'], $listenerArray['port']);

			if (isset($listenerArray['vhostMapList'])
					&& is_array($listenerArray['vhostMapList']))
			{
				$vhmapList = &$listenerArray['vhostMapList'];
				for ($k = 0; $k < count($vhmapList); ++$k)
				{

					$vh = array_column($vhmapList[$k][self::XML_ELEM],
							self::XML_VAL, self::XML_NAME);

					if (isset($vh['vhost']) && isset($vh['domain']))
					{
						$listener->appendVHMap($vh['vhost'], $vh['domain']);
					}
				}
				unset($listenerArray['vhostMapList']);
			}
			if (isset($listenerArray['secure']) && $listenerArray['secure'])
				$listener->importLSWSSsl($listenerArray);

			if (!isset($holder['listeners'])) {
				$holder['listeners'] = array();
			}
			$holder['listeners'][$listenerArray['name']] = &$listener;
		}
	}

	/**
	 * Extract the VHost configurations from the XML config.
	 *
	 * @param array $vhostList Array of VHost configurations.
	 * @param array $holder The destination array to put the extracted configurations.
	 */
	function extractVHosts(&$vhostList, &$holder)
	{
		for ($i = 0; $i < count($vhostList); ++$i)
		{
			$vh = array_column($vhostList[$i][self::XML_ELEM],
					self::XML_VAL, self::XML_NAME);
			if (!isset($vh['name']))
				continue;
			if (!isset($holder['vhTop'])) {
				$holder['vhTop'] = array();
			}
			$holder['vhTop'][$vh['name']] = &$vh;
		}
	}

	/**
	 * Write the vhost configuration file.
	 *
	 * @param string $vhname VHost name.
	 * @param array $vh VHost configuration.
	 * @param string $confFile The file name to write to.
	 * @return bool false on failure, true on success.
	 */
	function saveVhFile($vhname, &$vh, $confFile)
	{
		$fd = fopen($confFile, 'w');
		if (!$fd)
			return false;

		fputs($fd, '<?xml version="1.0" encoding="UTF-8"?>' . "\n");
		$level = 0;
		fputs($fd, $this->xmlTagS($level, 'virtualHostConfig'));

		//general
		$g = &$vh['general'];
		////docRoot
		fputs($fd, $this->xmlTag($level, 'docRoot', $g['docRoot']));
		fputs($fd, $this->xmlTag($level, 'enableGzip', $g['enableGzip']));
		fputs($fd, $this->xmlTag($level, 'adminEmails', $g['adminEmails']));

		////logging
		fputs($fd, $this->xmlTagS($level, 'logging'));
		if (isset($g['log_fileName']))
		{
			fputs($fd, $this->xmlTagS($level, 'log'));
			fputs($fd, $this->xmlTag($level, 'useServer', $g['log_useServer']));
			fputs($fd, $this->xmlTag($level, 'fileName', $g['log_fileName']));
			fputs($fd, $this->xmlTag($level, 'logLevel', $g['log_logLevel']));
			fputs($fd, $this->xmlTag($level, 'rollingSize', $g['log_rollingSize']));
			fputs($fd, $this->xmlTagE($level, 'log'));
		}
		if (isset($g['accessLog_fileName']))
		{
			fputs($fd, $this->xmlTagS($level, 'accessLog'));
			fputs($fd, $this->xmlTag($level, 'useServer', $g['accessLog_useServer']));
			fputs($fd, $this->xmlTag($level, 'fileName', $g['accessLog_fileName']));
			fputs($fd, $this->xmlTag($level, 'logReferer', $g['accessLog_logReferer']));
			fputs($fd, $this->xmlTag($level, 'logUserAgent', $g['accessLog_logUserAgent']));
			fputs($fd, $this->xmlTag($level, 'rollingSize', $g['accessLog_rollingSize']));
			fputs($fd, $this->xmlTag($level, 'keepDays', $g['accessLog_keepDays']));
			fputs($fd, $this->xmlTag($level, 'compressArchive', $g['accessLog_compressArchive']));
			fputs($fd, $this->xmlTagE($level, 'accessLog'));
		}
		fputs($fd, $this->xmlTagE($level, 'logging'));

		////index files
		$this->writeSection($fd, $level, $g['index'], 'index');

		//htAccess
		if (isset($g['allowOverride']) || isset($g['accessFileName']))
		{
			fputs($fd, $this->xmlTagS($level, 'htAccess'));
			fputs($fd, $this->xmlTag($level, 'allowOverride', $g['allowOverride']));
			fputs($fd, $this->xmlTag($level, 'accessFileName', $g['accessFileName']));
			fputs($fd, $this->xmlTagE($level, 'htAccess'));
		}

		//expires
		fputs($fd, $this->xmlTagS($level, 'expires'));
		fputs($fd, $this->xmlTag($level, 'enableExpires', $g['enableExpires']));
		fputs($fd, $this->xmlTag($level, 'expiresDefault', $g['expiresDefault']));
		fputs($fd, $this->xmlTagE($level, 'expires'));

		//security
		if (isset($vh['security']))
		{
			$s = &$vh['security'];
			fputs($fd, $this->xmlTagS($level, 'security'));
			fputs($fd, $this->xmlTagS($level, 'general'));
			fputs($fd, $this->xmlTag($level, 'enableContextAC', $s['general']['enableContextAC']));
			fputs($fd, $this->xmlTagE($level, 'general'));
			//// hotlink
			$this->writeSection($fd, $level, $s['hotlink'], 'hotlinkCtrl');

			//// realmList
			$keys = $this->loadKeys($s['realm']);
			if ($keys != NULL)
			{
				fputs($fd, $this->xmlTagS($level, 'realmList'));
				sort($keys);
				foreach ($keys as $key)
				{
					$name = $key;
					$realm = $s['realm'][$key];
					fputs($fd, $this->xmlTagS($level, 'realm'));
					fputs($fd, $this->xmlTag($level, 'name', $name));
					fputs($fd, $this->xmlTag($level, 'type', $realm['type']));
					fputs($fd, $this->xmlTagS($level, 'userDB'));
					fputs($fd, $this->xmlTag($level, 'location', $realm['userDB_location']));
					if ($realm['type'] == 'LDAP')
					{
						fputs($fd, $this->xmlTag($level, 'attrPasswd', $realm['userDB_attrPasswd']));
						fputs($fd, $this->xmlTag($level, 'attrMemberOf', $realm['userDB_attrMemberOf']));
					}
					fputs($fd, $this->xmlTag($level, 'maxCacheSize', $realm['userDB_maxCacheSize']));
					fputs($fd, $this->xmlTag($level, 'cacheTimeout', $realm['userDB_cacheTimeout']));
					fputs($fd, $this->xmlTagE($level, 'userDB'));
					fputs($fd, $this->xmlTagS($level, 'groupDB'));
					fputs($fd, $this->xmlTag($level, 'location', $realm['groupDB_location']));
					if ($realm['type'] == 'LDAP')
					{
						fputs($fd, $this->xmlTag($level, 'attrGroupMember', $realm['groupDB_attrGroupMember']));
					}
					fputs($fd, $this->xmlTag($level, 'maxCacheSize', $realm['groupDB_maxCacheSize']));
					fputs($fd, $this->xmlTag($level, 'cacheTimeout', $realm['groupDB_cacheTimeout']));
					fputs($fd, $this->xmlTagE($level, 'groupDB'));
					if ($realm['type'] == 'LDAP')
					{
						fputs($fd, $this->xmlTag($level, 'LDAPBindDN', $realm['LDAPBindDN']));
						fputs($fd, $this->xmlTag($level, 'LDAPBindPasswd', $realm['LDAPBindPasswd']));
					}
					fputs($fd, $this->xmlTagE($level, 'realm'));
				}
				fputs($fd, $this->xmlTagE($level, 'realmList'));
			}

			//// access control
			if ($s['accessControl']['accessControl_allow'] != NULL || $s['accessControl']['accessControl_deny'] != NULL)
			{
				fputs($fd, $this->xmlTagS($level, 'accessControl'));
				fputs($fd, $this->xmlTag($level, 'allow', $s['accessControl']['accessControl_allow']));
				fputs($fd, $this->xmlTag($level, 'deny', $s['accessControl']['accessControl_deny']));
				fputs($fd, $this->xmlTagE($level, 'accessControl'));
			}
			fputs($fd, $this->xmlTagE($level, 'security'));
		}

		//ext
		if (isset($vh['ext']) && count($vh['ext']) > 0)
		{
			fputs($fd, $this->xmlTagS($level, 'extProcessorList'));
			$ext0 = &$vh['ext'];
			$ext = array();
			foreach ($ext0 as $a)
			{
				$id = $a['type'] . $a['name'];
				$ext[$id] = $a;
			}
			ksort($ext);
			reset($ext);

			foreach ($ext as $f)
			{
				fputs($fd, $this->xmlTagS($level, 'extProcessor'));
				fputs($fd, $this->xmlTag($level, 'name', $f['name']));
				fputs($fd, $this->xmlTag($level, 'address', $f['address']));
				fputs($fd, $this->xmlTag($level, 'type', $f['type']));
				fputs($fd, $this->xmlTag($level, 'maxConns', $f['maxConns']));
				fputs($fd, $this->xmlTag($level, 'initTimeout', $f['initTimeout']));
				fputs($fd, $this->xmlTag($level, 'retryTimeout', $f['retryTimeout']));
				fputs($fd, $this->xmlTag($level, 'respBuffer', $f['respBuffer']));
				if (isset($f['env']))
				{
					foreach ($f['env'] as $value)
					{
						fputs($fd, $this->xmlTag($level, 'env', $value));
					}
				}
				if ($f['type'] == 'fcgi' || $f['type'] == 'fcgiauth')
				{
					fputs($fd, $this->xmlTag($level, 'autoStart', $f['autoStart']));
					fputs($fd, $this->xmlTag($level, 'path', $f['path']));
					fputs($fd, $this->xmlTag($level, 'backlog', $f['backlog']));
					fputs($fd, $this->xmlTag($level, 'instances', $f['instances']));
					fputs($fd, $this->xmlTag($level, 'priority', $f['priority']));
				}
				fputs($fd, $this->xmlTagE($level, 'extProcessor'));
			}
			fputs($fd, $this->xmlTagE($level, 'extProcessorList'));
		}

		//contexts
		$htypes = $this->loadKeys($vh['context']);
		if ($htypes != NULL)
		{
			fputs($fd, $this->xmlTagS($level, 'contextList'));
			foreach ($htypes as $htype)
			{
				$ct = &$vh['context'][$htype];
				foreach ($ct as $ctx)
				{
					fputs($fd, $this->xmlTagS($level, 'context'));

					foreach ($ctx as $key => $val)
					{
						if ($val !== '' && strncmp($key, 'accessControl', 13) != 0 && strncmp($key, 'rewrite', 7) != 0 && ($key != 'authorizer'))
							fputs($fd, $this->xmlTag($level, $key, $val));
					}
					//// access control
					if ($ctx['accessControl_allow'] != NULL || $ctx['accessControl_deny'] != NULL)
					{
						fputs($fd, $this->xmlTagS($level, 'accessControl'));
						fputs($fd, $this->xmlTag($level, 'allow', $ctx['accessControl_allow']));
						fputs($fd, $this->xmlTag($level, 'deny', $ctx['accessControl_deny']));
						fputs($fd, $this->xmlTagE($level, 'accessControl'));
					}
					if ($ctx['authorizer'] != NULL)
					{
						fputs($fd, $this->xmlTagS($level, 'extAuthorizer'));
						fputs($fd, $this->xmlTag($level, 'type', 'fcgiauth'));
						fputs($fd, $this->xmlTag($level, 'handler', $ctx['authorizer']));
						fputs($fd, $this->xmlTagE($level, 'extAuthorizer'));
					}
					if ($ctx['rewrite_enable'] != NULL || $ctx['rewrite_inherit'] != NULL || $ctx['rewrite_base'] != NULL || $ctx['rewrite_rules'] != NULL)
					{
						fputs($fd, $this->xmlTagS($level, 'rewrite'));
						fputs($fd, $this->xmlTag($level, 'enable', $ctx['rewrite_enable']));
						fputs($fd, $this->xmlTag($level, 'inherit', $ctx['rewrite_inherit']));
						fputs($fd, $this->xmlTag($level, 'base', $ctx['rewrite_base']));
						fputs($fd, $this->xmlTag($level, 'rules', $ctx['rewrite_rules']));
						fputs($fd, $this->xmlTagE($level, 'rewrite'));
					}

					fputs($fd, $this->xmlTagE($level, 'context'));
				}
			}
			fputs($fd, $this->xmlTagE($level, 'contextList'));
		}

		//script handler
		fputs($fd, $this->xmlTagS($level, 'scriptHandlerList'));
		$keys = $this->loadKeys($vh['scriptHandler']);
		if ($keys != NULL)
		{
			$sc = &$vh['scriptHandler'];
			sort($keys);
			foreach ($keys as $key)
			{
				fputs($fd, $this->xmlTagS($level, 'scriptHandler'));
				fputs($fd, $this->xmlTag($level, 'suffix', $sc[$key]['suffix']));
				fputs($fd, $this->xmlTag($level, 'type', $sc[$key]['type']));
				if (isset($sc[$key]['handler']))
					fputs($fd, $this->xmlTag($level, 'handler', $sc[$key]['handler']));
				fputs($fd, $this->xmlTagE($level, 'scriptHandler'));
			}
		}
		fputs($fd, $this->xmlTagE($level, 'scriptHandlerList'));

		//error pages
		$keys = $this->loadKeys($vh['errUrl']);
		if ($keys != NULL)
		{
			fputs($fd, $this->xmlTagS($level, 'customErrorPages'));
			sort($keys);
			foreach ($keys as $key)
			{
				fputs($fd, $this->xmlTagS($level, 'errorPage'));
				fputs($fd, $this->xmlTag($level, 'errCode', $key));
				fputs($fd, $this->xmlTag($level, 'url', $vh['errUrl'][$key]['url']));
				fputs($fd, $this->xmlTagE($level, 'errorPage'));
			}
			fputs($fd, $this->xmlTagE($level, 'customErrorPages'));
		}

		//rewrite
		fputs($fd, $this->xmlTagS($level, 'rewrite'));
		if (isset($vh['rewrite']['ctrl']))
		{
			fputs($fd, $this->xmlTag($level, 'enable', $vh['rewrite']['ctrl']['enable']));
			fputs($fd, $this->xmlTag($level, 'logLevel', $vh['rewrite']['ctrl']['logLevel']));
		}
		$keys = $this->loadKeys($vh['rewrite']['map']);
		if ($keys != NULL)
		{
			sort($keys);
			foreach ($keys as $key)
			{
				fputs($fd, $this->xmlTagS($level, 'map'));
				fputs($fd, $this->xmlTag($level, 'name', $key));
				fputs($fd, $this->xmlTag($level, 'location', $vh['rewrite']['map'][$key]['location']));
				fputs($fd, $this->xmlTagE($level, 'map'));
			}
		}
		fputs($fd, $this->xmlTag($level, 'rules', $vh['rewrite']['rules']));
		fputs($fd, $this->xmlTagE($level, 'rewrite'));

		//ssl
		if ((isset($vh['ssl'])) // ssl is set AND (
				&& ((!isset($vh['ssl']['sslengine'])) // engine does not exist
					|| (($vh['ssl']['sslengine'])))) // OR engine is on)
		{
			$ssl = $vh['ssl'];
			fputs($fd, $this->xmlTagS($level, 'vhssl'));
			fputs($fd, $this->xmlTag($level, 'certFile', $ssl['sslcertificatefile']));
			if (isset($ssl['sslcertificatekeyfile']))
				fputs($fd, $this->xmlTag($level, 'keyFile', $ssl['sslcertificatekeyfile']));
			else
				fputs($fd, $this->xmlTag($level, 'keyFile', $ssl['sslcertificatefile']));
			if (isset($ssl['sslciphersuite']))
				fputs($fd, $this->xmlTag($level, 'ciphers', $ssl['sslciphersuite']));
			if (isset($ssl['sslcertificatechainfile']))
			{
				fputs($fd, $this->xmlTag($level, 'certChain', 1));
				fputs($fd, $this->xmlTag($level, 'CACertFile', $ssl['sslcertificatechainfile']));
			}
			fputs($fd, $this->xmlTagE($level, 'vhssl'));
		}

		//frontPage
		$this->writeSection($fd, $level, $vh['addons']['frontPage'], 'frontPage');

		//AWStats
		$this->writeSection($fd, $level, $vh['addons']['awstats'], 'awstats');

		fputs($fd, $this->xmlTagE($level, 'virtualHostConfig'));
		fclose($fd);
		clearstatcache();

		return true;
	}

	/**
	 * Write out the LSWS server configuration based on extracted configurations.
	 *
	 * @param array $serv The extracted server configurations
	 * @param string $filename The file to write to.
	 */
	function saveServerFile(&$serv, $filename)
	{
		$fd = fopen($filename, 'w');
		if (!$fd)
			return false;
		fputs($fd, '<?xml version="1.0" encoding="UTF-8"?>' . "\n");
		$level = 0;
		fputs($fd, $this->xmlTagS($level, 'httpServerConfig'));

		$g = &$serv['general'];
		fputs($fd, $this->xmlTag($level, 'serverName', $g['serverName']));
		fputs($fd, $this->xmlTag($level, 'user', $g['user']));
		fputs($fd, $this->xmlTag($level, 'group', $g['group']));
		fputs($fd, $this->xmlTag($level, 'priority', $g['priority']));
		fputs($fd, $this->xmlTag($level, 'chrootPath', $g['chrootPath']));
		fputs($fd, $this->xmlTag($level, 'enableChroot', $g['enableChroot']));
		fputs($fd, $this->xmlTag($level, 'inMemBufSize', $g['inMemBufSize']));
		fputs($fd, $this->xmlTag($level, 'swappingDir', $g['swappingDir']));
		fputs($fd, $this->xmlTag($level, 'autoRestart', $g['autoRestart']));
		fputs($fd, $this->xmlTag($level, 'adminRoot', $g['adminRoot']));

		fputs($fd, $this->xmlTag($level, 'mime', $g['mime']));
		if (isset($g['indexFiles']))
			fputs($fd, $this->xmlTag($level, 'indexFiles', $g['indexFiles']));
		fputs($fd, $this->xmlTag($level, 'showVersionNumber', $g['showVersionNumber']));
		fputs($fd, $this->xmlTag($level, 'adminEmails', $g['adminEmails']));

		// logging
		fputs($fd, $this->xmlTagS($level, 'logging'));
		fputs($fd, $this->xmlTagS($level, 'log'));
		fputs($fd, $this->xmlTag($level, 'fileName', $g['log_fileName']));
		fputs($fd, $this->xmlTag($level, 'logLevel', $g['log_logLevel']));
		fputs($fd, $this->xmlTag($level, 'debugLevel', $g['log_debugLevel']));
		fputs($fd, $this->xmlTag($level, 'rollingSize', $g['log_rollingSize']));
		fputs($fd, $this->xmlTag($level, 'enableStderrLog', $g['log_enableStderrLog']));
		fputs($fd, $this->xmlTagE($level, 'log'));
		fputs($fd, $this->xmlTagS($level, 'accessLog'));
		fputs($fd, $this->xmlTag($level, 'fileName', $g['accessLog_fileName']));
		fputs($fd, $this->xmlTag($level, 'logReferer', $g['accessLog_logReferer']));
		fputs($fd, $this->xmlTag($level, 'logUserAgent', $g['accessLog_logUserAgent']));
		fputs($fd, $this->xmlTag($level, 'rollingSize', $g['accessLog_rollingSize']));
		fputs($fd, $this->xmlTag($level, 'keepDays', $g['accessLog_keepDays']));
		fputs($fd, $this->xmlTag($level, 'compressArchive', $g['accessLog_compressArchive']));
		fputs($fd, $this->xmlTagE($level, 'accessLog'));
		fputs($fd, $this->xmlTagE($level, 'logging'));


		//htAccess
		if (isset($g['allowOverride']) || isset($g['accessFileName']))
		{
			fputs($fd, $this->xmlTagS($level, 'htAccess'));
			fputs($fd, $this->xmlTag($level, 'allowOverride', $g['allowOverride']));
			fputs($fd, $this->xmlTag($level, 'accessFileName', $g['accessFileName']));
			fputs($fd, $this->xmlTagE($level, 'htAccess'));
		}

		//expires
		fputs($fd, $this->xmlTagS($level, 'expires'));
		fputs($fd, $this->xmlTag($level, 'enableExpires', $g['enableExpires']));
		fputs($fd, $this->xmlTag($level, 'expiresDefault', $g['expiresDefault']));
		fputs($fd, $this->xmlTag($level, 'expiresByType', $g['expiresByType']));
		fputs($fd, $this->xmlTagE($level, 'expires'));

		//tuning
		$this->writeSection($fd, $level, $serv['tuning'], 'tuning');

		//security
		$s = &$serv['security'];
		fputs($fd, $this->xmlTagS($level, 'security'));

		fputs($fd, $this->xmlTagS($level, 'fileAccessControl'));
		fputs($fd, $this->xmlTag($level, 'followSymbolLink', $s['fileAccessControl_followSymbolLink']));
		fputs($fd, $this->xmlTag($level, 'checkSymbolLink', $s['fileAccessControl_checkSymbolLink']));
		fputs($fd, $this->xmlTag($level, 'requiredPermissionMask', $s['fileAccessControl_requiredPermissionMask']));
		fputs($fd, $this->xmlTag($level, 'restrictedPermissionMask', $s['fileAccessControl_restrictedPermissionMask']));
		fputs($fd, $this->xmlTagE($level, 'fileAccessControl'));

		fputs($fd, $this->xmlTagS($level, 'CGIRLimit'));
		fputs($fd, $this->xmlTag($level, 'maxCGIInstances', $s['CGIRLimit_maxCGIInstances']));
		fputs($fd, $this->xmlTag($level, 'minUID', $s['CGIRLimit_minUID']));
		fputs($fd, $this->xmlTag($level, 'minGID', $s['CGIRLimit_minGID']));
		fputs($fd, $this->xmlTag($level, 'priority', $s['CGIRLimit_priority']));
		fputs($fd, $this->xmlTag($level, 'CPUSoftLimit', $s['CGIRLimit_CPUSoftLimit']));
		fputs($fd, $this->xmlTag($level, 'CPUHardLimit', $s['CGIRLimit_CPUHardLimit']));
		fputs($fd, $this->xmlTag($level, 'memSoftLimit', $s['CGIRLimit_memSoftLimit']));
		fputs($fd, $this->xmlTag($level, 'memHardLimit', $s['CGIRLimit_memHardLimit']));
		fputs($fd, $this->xmlTag($level, 'procSoftLimit', $s['CGIRLimit_procSoftLimit']));
		fputs($fd, $this->xmlTag($level, 'procHardLimit', $s['CGIRLimit_procHardLimit']));
		fputs($fd, $this->xmlTagE($level, 'CGIRLimit'));

		fputs($fd, $this->xmlTagS($level, 'perClientConnLimit'));
		fputs($fd, $this->xmlTag($level, 'staticReqPerSec', $s['perClientConnLimit_staticReqPerSec']));
		fputs($fd, $this->xmlTag($level, 'dynReqPerSec', $s['perClientConnLimit_dynReqPerSec']));
		fputs($fd, $this->xmlTag($level, 'outBandwidth', $s['perClientConnLimit_outBandwidth']));
		fputs($fd, $this->xmlTag($level, 'inBandwidth', $s['perClientConnLimit_inBandwidth']));
		fputs($fd, $this->xmlTag($level, 'softLimit', $s['perClientConnLimit_softLimit']));
		fputs($fd, $this->xmlTag($level, 'hardLimit', $s['perClientConnLimit_hardLimit']));
		fputs($fd, $this->xmlTag($level, 'gracePeriod', $s['perClientConnLimit_gracePeriod']));
		fputs($fd, $this->xmlTag($level, 'banPeriod', $s['perClientConnLimit_banPeriod']));
		fputs($fd, $this->xmlTagE($level, 'perClientConnLimit'));

		fputs($fd, $this->xmlTagS($level, 'accessDenyDir'));
		if (isset($s['accessDenyDir']))
		{
			foreach ($s['accessDenyDir'] as $value)
			{
				fputs($fd, $this->xmlTag($level, 'dir', $value));
			}
		}
		fputs($fd, $this->xmlTagE($level, 'accessDenyDir'));
		fputs($fd, $this->xmlTagS($level, 'accessControl'));
		if (isset($s['accessControl_allow']))
			fputs($fd, $this->xmlTag($level, 'allow', $s['accessControl_allow']));
		if (isset($s['accessControl_deny']))
			fputs($fd, $this->xmlTag($level, 'deny', $s['accessControl_deny']));
		fputs($fd, $this->xmlTagE($level, 'accessControl'));
		fputs($fd, $this->xmlTagE($level, 'security'));

		//ext
		fputs($fd, $this->xmlTagS($level, 'extProcessorList'));
		if (isset($serv['ext']) && count($serv['ext']) > 0)
		{
			$ext0 = &$serv['ext'];
			$ext = array();
			foreach ($ext0 as $a)
			{
				$id = $a['type'] . $a['name'];
				$ext[$id] = $a;
			}
			ksort($ext);
			reset($ext);

			foreach ($ext as $f)
			{
				fputs($fd, $this->xmlTagS($level, 'extProcessor'));
				fputs($fd, $this->xmlTag($level, 'name', $f['name']));
				fputs($fd, $this->xmlTag($level, 'address', $f['address']));
				fputs($fd, $this->xmlTag($level, 'type', $f['type']));
				fputs($fd, $this->xmlTag($level, 'maxConns', $f['maxConns']));
				fputs($fd, $this->xmlTag($level, 'initTimeout', $f['initTimeout']));
				fputs($fd, $this->xmlTag($level, 'retryTimeout', $f['retryTimeout']));
				fputs($fd, $this->xmlTag($level, 'respBuffer', $f['respBuffer']));
				if (isset($f['env']))
				{
					foreach ($f['env'] as $value)
					{
						fputs($fd, $this->xmlTag($level, 'env', $value));
					}
				}
				if ($f['type'] == 'fcgi' || $f['type'] == 'fcgiauth' || $f['type'] == 'lsapi' || $f['type'] == 'logger')
				{
					fputs($fd, $this->xmlTag($level, 'autoStart', $f['autoStart']));
					fputs($fd, $this->xmlTag($level, 'path', $f['path']));
					fputs($fd, $this->xmlTag($level, 'backlog', $f['backlog']));
					fputs($fd, $this->xmlTag($level, 'instances', $f['instances']));
					fputs($fd, $this->xmlTag($level, 'priority', $f['priority']));
				}
				fputs($fd, $this->xmlTagE($level, 'extProcessor'));
			}
		}
		fputs($fd, $this->xmlTagE($level, 'extProcessorList'));

		//script handler
		fputs($fd, $this->xmlTagS($level, 'scriptHandlerList'));
		$keys = $this->loadKeys($serv['scriptHandler']);
		if ($keys != NULL)
		{
			$sc = &$serv['scriptHandler'];
			sort($keys);
			foreach ($keys as $key)
			{
				fputs($fd, $this->xmlTagS($level, 'scriptHandler'));
				fputs($fd, $this->xmlTag($level, 'suffix', $sc[$key]['suffix']));
				fputs($fd, $this->xmlTag($level, 'type', $sc[$key]['type']));
				if (isset($sc[$key]['handler']))
					fputs($fd, $this->xmlTag($level, 'handler', $sc[$key]['handler']));
				fputs($fd, $this->xmlTagE($level, 'scriptHandler'));
			}
		}
		fputs($fd, $this->xmlTagE($level, 'scriptHandlerList'));

		//listner
		fputs($fd, $this->xmlTagS($level, 'listenerList'));
		$lnames = array_keys($serv['listeners']);
		sort($lnames);
		reset($lnames);
		foreach ($lnames as $ln)
		{
			$l = &$serv['listeners'][$ln];
			fputs($fd, $this->xmlTagS($level, 'listener'));
			fputs($fd, $this->xmlTag($level, 'name', $ln));
			$ip = $l->getIp();
			if ($ip == 'ANY')
				$ip = '*';
			$addr = $ip . ':' . $l->getPort();
			fputs($fd, $this->xmlTag($level, 'address', $addr));
			fputs($fd, $this->xmlTag($level, 'binding', $l->getBinding()));
			fputs($fd, $this->xmlTag($level, 'secure', $l->isSecure()));
			if ($l->isSecure())
			{
				fputs($fd, $this->xmlTag($level, 'certFile', $l->getSslCert()));
				fputs($fd, $this->xmlTag($level, 'keyFile', $l->getSslKey()));
				fputs($fd, $this->xmlTag($level, 'CACertFile', $l->getSslCaBundle()));
				fputs($fd, $this->xmlTag($level, 'ciphers', $l->getSslCipher()));
			}
			fputs($fd, $this->xmlTagS($level, 'vhostMapList'));
			$map = $l->getVHMap();
			$keys = $this->loadKeys($map);
			if ($keys != NULL)
			{
				sort($keys);
				foreach ($keys as $key)
				{
					fputs($fd, $this->xmlTagS($level, 'vhostMap'));
					fputs($fd, $this->xmlTag($level, 'vhost', $key));
					fputs($fd, $this->xmlTag($level, 'domain', $map[$key]));
					fputs($fd, $this->xmlTagE($level, 'vhostMap'));
				}
			}
			fputs($fd, $this->xmlTagE($level, 'vhostMapList'));
			fputs($fd, $this->xmlTagE($level, 'listener'));
		}
		fputs($fd, $this->xmlTagE($level, 'listenerList'));

		//vh
		fputs($fd, $this->xmlTagS($level, 'virtualHostList'));
		$hnames = array_keys($serv['vhTop']);
		sort($hnames);
		foreach ($hnames as $hn)
		{
			if ($hn)
			{
				$vh = &$serv['vhTop'][$hn];
				if (substr($vh['vhRoot'], -1) != '/')
					$vh['vhRoot'] .= '/';
				$this->writeSection($fd, $level, $vh, 'virtualHost');
			}
		}
		fputs($fd, $this->xmlTagE($level, 'virtualHostList'));

		fputs($fd, $this->xmlTagE($level, 'httpServerConfig'));
		fclose($fd);

		return true;
	}

	function &loadKeys(&$list)
	{
		$keys = NULL;
		if (isset($list) && count($list) > 0)
			$keys = array_keys($list);

		return $keys;
	}

	function writeSection(&$fd, &$level, &$holder, $sectionName)
	{
		if (!isset($holder))
			return;
		fputs($fd, $this->xmlTagS($level, $sectionName));
		foreach ($holder as $key => $val)
		{
			fputs($fd, $this->xmlTag($level, $key, $val));
		}
		fputs($fd, $this->xmlTagE($level, $sectionName));
	}

	function xmlTag($level, $tag, $value)
	{
		if (is_array($value))
		{
			echo "tag=$tag\n";
			var_dump($value);
		}
		$val = htmlspecialchars($value);
		return str_repeat('  ', $level) . "<$tag>$val</$tag>\n";
	}

	function xmlTagS(&$level, $tag)
	{
		return str_repeat('  ', $level++) . "<$tag>\n";
	}

	function xmlTagE(&$level, $tag)
	{
		return str_repeat('  ', --$level) . "</$tag>\n";
	}

	function dbg_out($tag, &$obj)
	{
		echo "<!-- $tag --\n";
		var_dump($obj);
		echo "-->\n";
	}

	function dbg_tag($tag)
	{
		echo "<!-- $tag -->\n";
	}

}

?>
