<?php

/********************************************
* LiteSpeed AP2ZConf project
* @Author:   LiteSpeed Technologies, Inc. (http://www.litespeedtech.com)
* @Copyright: (c) 2017
*********************************************/

require_once('LSWSConf.php');
require_once('ParsedConf.php');
require_once('MigrUtil.php');
require_once('VHostConf.php');


class Extractor
{

	var $_apacheServerRoot;     # Used for relative path extraction.
	var $_serv = array();       # Combined conf.
	var $_vhosts = array();
	var $_vhnames = array();
	var $_debug = 0;

	/**
	 * Load and parse the Apache configurations.
	 *
	 * @param string $apacheServerRoot Absolute path to the main apache configurations path.
	 * @param string $lswsServerRoot Absolute path to LSWS server root.
	 */
	public function __construct($apacheServerRoot, $lswsServerRoot)
	{
		$this->_apacheServerRoot = $apacheServerRoot;
		// load default
		$lswsConf = new LSWSConf ();
		$this->_serv = $lswsConf->loadServConf($lswsServerRoot);
	}

	function extract(&$rawData)
	{
		$this->extractServGeneral($rawData->_configEntries);
		$this->extractServSecurity($rawData->_configEntries, $this->_serv ['security']);
		$this->extractTuning($rawData->_configEntries);
		$this->extractListener($rawData->_configEntries);
		$this->extractVhosts($rawData);
		// var_dump( $this->_serv['listeners'] );
	}

	/**
	 * Given Apache server conf, extract VHost configs and append ot VHostMap for each listener.
	 *
	 * @param Block $block The associated Block for the configurations.
	 */
	function extractVhosts(&$block)
	{
		$catchAll = array();
		$this->extractVH($block, 'ApacheMain', $catchAll);
		// var_dump( $block->_c['VirtualHost'] );
		if (isset($block->_childrenBlocks[Block::BT_VHOST]))
		{
			$vhnames = array_keys($block->_childrenBlocks [Block::BT_VHOST]);
			foreach ($vhnames as $name)
			{
				for ($i = 0; $i < count($block->_childrenBlocks [Block::BT_VHOST] [$name]); ++$i)
					$this->extractVH($block->_childrenBlocks [Block::BT_VHOST] [$name] [$i], $name, $catchAll);
			}
		}
		// analyse and set all other values,
		// add catchAll domain
		foreach ($catchAll as $addr => $vhname)
		{
			$listener = &$this->_serv['listeners'][$addr];
			$listener->appendVHMap($vhname, '*');
		}

		$vhnames = array_keys($this->_vhosts);
		foreach ($vhnames as $name)
		{
			$this->_serv ['vhTop'] [$name] = $this->_vhosts [$name] ['vhTop'];
		}
	}

	/**
	 * Extract an Apache VHost configuration.
	 *
	 * @param Block $block The virtual host block.
	 * @param string $name The Apache VHost name - typically ip:port
	 * @param array $catchAll Forwarded to addVhMap
	 */
	function extractVH(&$block, $name, &$catchAll)
	{
		$vh = array(
			'vhTop' => array(),
			'general' => array(),
		);
		$vh ['vhTop'] ['name'] = $name;
		$domains = array();
		$vhname = '';
		$byType = array();
		$allow = null;
		$deny = null;

		$list = &$block->_configEntries;
		for ($i = 0; $i < count($list); ++$i)
		{
			$entry = &$list [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			if ($entry->processTag('ServerName'))
			{
				if ($vhname == '')
				{
					$vhname = $entry->_val;
					$j = 0;
					while (isset($this->_vhnames[$vhname]))
					{
						++$j;
						$vhname = $entry->_val . $j;
					}
					$this->_vhnames [$vhname] = 1;
					$vh ['vhTop'] ['name'] = $vhname;
					$domains [] = $entry->_val;
				}
			}
			elseif ($entry->processTag('ServerAlias'))
			{
				$alias = preg_split("/\s+/", $entry->_val);
				if ($alias !== false) {
					$domains = array_merge($domains, $alias);
				}
			}
			elseif ($entry->processTag('DocumentRoot'))
			{
				$vh ['general'] ['docRoot'] = $entry->_val;
				$vh ['vhTop'] ['vhRoot'] = $entry->_val;
			}
			elseif ($entry->processTag('DirectoryIndex'))
			{
				if (!isset($vh ['general'] ['index'])) {
					$vh ['general'] ['index'] = array();
				}
				$vh ['general'] ['index'] ['indexFiles'] = MigrUtil::getCommaList($entry->_val);
			}
			elseif ($entry->processTag('Options'))
			{
				$res = $this->containOption($entry->_val, 'Indexes');
				if ($res != - 1)
				{
					if (!isset($vh ['general'] ['index'])) {
						$vh ['general'] ['index'] = array();
					}
					$vh ['general'] ['index'] ['autoIndex'] = $res;
				}
				$res = $this->containOption($entry->_val, 'FollowSymLinks');
				if ($res != - 1)
					$vh ['vhTop'] ['allowSymbolLink'] = $res;
				$res = $this->containOption($entry->_val, 'SymLinksIfOwnerMatch');
				if ($res == 1)
					$vh ['vhTop'] ['allowSymbolLink'] = 2;
			}
			elseif ($entry->processTag('Order'))
			{
				$order = $entry->_val;
			}
			elseif ($entry->processTag('Allow'))
			{
				$allow = $entry->_val;
			}
			elseif ($entry->processTag('Deny'))
			{
				$deny = $entry->_val;
			}
			elseif ($entry->processTag('zctemplate'))
			{
				$vh ['template'] = $entry->_val;
			}
			elseif ($entry->processTag('zccluster'))
			{
				$vh ['cluster'] = $entry->_val;
			}
			elseif (( $entry->processTag('SSLCertificateFile') ) ||
					( $entry->processTag('SSLCertificateKeyFile') ) ||
					( $entry->processTag('SSLCertificateChainFile') ) ||
					( $entry->processTag('SSLCipherSuite') ) ||
					( $entry->processTag('SSLEngine') ))
			{
				$vh ['ssl'] [strtolower($entry->_directive)] = $entry->_val;
			}
			else
				$this->extractExpires($entry, $vh ['general'], $byType);
		}

		$vh ['domains'] = array_unique($domains);
		$this->addVhMap($vh, $vhname, $name, $catchAll);

		$this->extractLogging($list, $vh ['general']);
		$vh ['general'] ['log_useServer'] = (isset($vh ['general'] ['log_fileName'])) ? 0 : 1;
		$vh ['general'] ['accessLog_useServer'] = (isset($vh ['general'] ['accessLog_fileName'])) ? 0 : 1;

		if (!empty($byType))
			$vh ['general'] ['expiresByType'] = implode(',', $byType);

		if (isset($order))
			$this->extractAccessControl($order, $allow, $deny, $vh ['security'] ['accessControl']);

		$this->extractVhRewrite($list, $vh);
		$this->extractContext($block, $vh);
		$vh ['vhTop'] ['configFile'] = '$SERVER_ROOT/conf/' . $vhname . '.xml';

		$vh ['apName'] = $block->_name;

		$this->_vhosts [$vhname] = $vh;
	}

	/**
	 * Check for SSL configurations in the vhost conf and apply them in
	 * the VHostConf.
	 *
	 * @param array $vh The vhost configuration to extract from.
	 * @param VHostConf $l The listener to apply the configuration to.
	 */
	function configssl(&$vh, &$l)
	{

		if (!isset($vh ['ssl'] ['sslengine']))
			return;
		if (strcasecmp($vh ['ssl'] ['sslengine'], 'on'))
			return;
		if (!isset($vh['ssl']['sslcertificatefile']))
			return;
		$l->importVHSsl($vh ['ssl']);
	}

	/**
	 * Extract an Apache VHost configuration.
	 *
	 * @param array $vh The configuration array for the VHost to add.
	 * @param string $vhname The vhost name.
	 * @param string $name The Apache VHost name (generally the listener)
	 * @param array $catchAll Associated Array to set the current VHost as catch all.
	 */
	function addVhMap(&$vh, &$vhname, $name, &$catchAll)
	{
		$laddrs = array_keys($this->_serv ['listeners']);
		if ($vhname == '')
			$vhname = $name;

		if ($name == 'ApacheMain')
		{ //default server
			return;
		}

		$tmpvn = preg_split("/\s+/", $name);
		if ($tmpvn === false) {
			return;
		}
		foreach ($tmpvn as $vn)
		{
			$pos = strpos($vn, ':');
			if ($pos === false)
			{
				$ip = $vn;
				$port = '*';
			}
			else
			{
				$ip = substr($vn, 0, $pos);
				$port = substr($vn, $pos + 1);
			}

			foreach ($laddrs as $addr)
			{
				$l = &$this->_serv ['listeners'] [$addr];
				if (($l->isNameVH()) &&
						(!isset($catchAll[$addr])))
					$catchAll [$addr] = "ApacheMain";
				if ($this->checkVhListenerMap($ip, $port, $l->getIp(), $l->getPort()))
				{
					$l->appendVHMap($vhname, implode(',', $vh ['domains']));
					if (( $ip == '_default_' ) || (!$l->isNameVH()))
					{
						$catchAll [$addr] = $vhname;
						if (isset($vh['ssl']))
						{
							$this->configssl($vh, $l);
						}
					}
					elseif ($ip != '*')
						$l->appendVHMap($vhname, $ip);
				}
			}
		}
	}

	/**
	 * Compare VHost IP and port with listener IP and port. Return true if match.
	 *
	 * @param string $vh_ip The VHost IP to compare.
	 * @param string $vh_port The VHost Port to compare
	 * @param string $l_ip The listener IP to compare.
	 * @param string $l_port The listener port to compare.
	 * @return bool True if match, false otherwise.
	 */
	function checkVhListenerMap($vh_ip, $vh_port, $l_ip, $l_port)
	{
		if ($vh_port != '*')
		{
			if ($vh_port != $l_port)
				return false;
		}
		if (($l_ip == '*') || ($vh_ip == '*') || ($vh_ip == '_default_') || ($l_ip == $vh_ip))
		{
			if ($this->_debug)
				echo "checkVhListenerMap=true VH $vh_ip:$vh_port -- VHostConf $l_ip:$l_port \n";
			return true;
		}
		else
			return false;
	}

	/**
	 * Extract Tuning Configurations.
	 *
	 * @param array $entryList Array of Entry objects.
	 */
	function extractTuning(&$entryList)
	{
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;

			if ($entry->processTag('Timeout'))
			{
				$this->_serv ['tuning'] ['connTimeout'] = $entry->_val;
			}
		}
	}

	/**
	 * Extract Logging Configurations from the parsed Apache Configs.
	 *
	 * $holder contains the already-merged configurations.
	 *
	 * @param array $entryList Array of Entry objects.
	 * @param array $holder The destination array to extract to.
	 */
	function extractLogging(&$entryList, &$holder)
	{
		$hasCustomLog = false;
		$holder ['log_logLevel'] = 'INFO';
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;

			if ($entry->processTag('ErrorLog'))
			{
				if ($entry->_val[0] == '|')
				{
					$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Does not allow pipe, use default value instead.');
				}
				elseif (strncmp($entry->_val, 'syslog', 6) == 0)
				{
					$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Does not support syslog, use default value instead.');
				}
				else
				{
					$log_fileName = $entry->_val;
					if ($log_fileName[0] != '/')
						$holder ['log_fileName'] = $this->_apacheServerRoot . '/' . $log_fileName;
					else
						$holder ['log_fileName'] = $log_fileName;
				}
			}
			elseif ($entry->processTag('LogLevel'))
			{
				$levels = array('error', 'warn', 'notice', 'info', 'debug');
				if (in_array($entry->_val, $levels))
					$holder ['log_logLevel'] = strtoupper($entry->_val);
			}
			elseif ($entry->processTag('TransferLog'))
			{
				if ($entry->_val[0] == '|')
				{
					$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Does not allow pipe, use default value instead.');
				}
				else
				{
					$accessLog_fileName = $entry->_val;
					if ($accessLog_fileName[0] != '/')
						$holder ['accessLog_fileName'] = $this->_apacheServerRoot . '/' . $accessLog_fileName;
					else
						$holder ['accessLog_fileName'] = $accessLog_fileName;
				}
			}
			elseif ($entry->processTag('CustomLog'))
			{
				$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Does not support customized log.');
				$hasCustomLog = true;
			}
		}
		// if ( $hasCustomLog && !isset($holder['accessLog_fileName']) )
		// $holder['accessLog_fileName'] = 'logs/access_log';
	}

	/**
	 * Extract Expires configurations
	 *
	 * If the Entry is an ExpiresByType tag, the value is appended to the
	 * byType array.
	 *
	 * @param Entry $entry The corresponding configuration Entry.
	 * @param array $holder The destination array to extract to.
	 * @param array $byType Array of types to expire by.
	 */
	function extractExpires(&$entry, &$holder, &$byType)
	{
		// ('enableExpires', 'expiresByType', 'expiresDefault');
		if ($entry->processTag('ExpiresActive'))
		{
			$holder ['enableExpires'] = MigrUtil::getBool($entry->_val);
		}
		elseif ($entry->processTag('ExpiresDefault'))
		{
			$holder ['expiresDefault'] = $entry->_val;
		}
		elseif ($entry->processTag('ExpiresByType'))
		{
			$byType [] = $entry->_val;
		}
	}

	/**
	 * Create a context for an alias configuration.
	 *
	 * @param Entry $entry The alias configuration
	 * @param array $holder The destination array to extract to.
	 */
	function newAliasContext(&$entry, &$holder)
	{
		$isRegEx = (stristr($entry->_directive, 'Match') !== false);
		$type = (stristr($entry->_directive, 'Script') === false) ? 'NULL' : 'cgi';

		$l = MigrUtil::splitToList($entry->_val);
		if (count($l) == 2)
		{
			$ctx = array();
			$ctx ['type'] = $type;
			if ($isRegEx)
				$ctx ['uri'] = 'exp:' . $l [0];
			else
				$ctx ['uri'] = $l [0];
			$ctx ['location'] = $l [1];
			$holder [$ctx ['uri']] = $ctx;
		}
		else
			$entry->setStatus(MigrUtil::STATUS_ERROR, 'Wrong parameters');
	}

	/**
	 * Create a context for a redirect configuration.
	 *
	 * @param Entry $entry The redirect configuration
	 * @param array $holder The destination array to extract to.
	 */
	function newRedirectContext(&$entry, &$holder)
	{
		$l = MigrUtil::splitToList($entry->_val);
		$isRegEx = (stristr($entry->_directive, 'Match') !== false);


		$codes = array('permanent' => 301, 'temp' => 302, 'seeother' => 303, 'gone' => 410);
		$ctx = array();
		$ctx ['type'] = 'redirect';

		if (count($l) == 2)
		{
			$uri = $l [0];
			$loc = $l [1];
		}
		elseif (count($l) == 3)
		{
			$status = $l [0];
			if (!is_numeric($status))
			{
				if (isset($codes [$status]))
					$status = $codes [$status];
				else
					$status = 302;
			}
			$ctx ['statusCode'] = $status;
			$uri = $l [1];
			$loc = $l [2];
		}
		else
		{
			$entry->setStatus(MigrUtil::STATUS_ERROR, 'Wrong parameters');
			return;
		}

		if (strcasecmp($entry->_directive, 'RedirectTemp') == 0)
			$status = 302;
		elseif (strcasecmp($entry->_directive, 'RedirectPermanent') == 0)
			$status = 301;

		$ctx ['externalRedirect'] = ($uri [0] == '/' && !isset($status)) ? 0 : 1;

		if ($isRegEx)
			$ctx ['uri'] = 'exp:' . $uri;
		else
			$ctx ['uri'] = $uri;
		$ctx ['location'] = $loc;
		$holder [$ctx ['uri']] = $ctx;
	}

	/**
	 * Check the Entry list to see if there are tags which should be converted to Context.
	 *
	 * Alias and Redirect tags should be converted to context.
	 *
	 * @param array $entryList Array of Entry objects.
	 * @param array $holder The destination array to extract to.
	 */
	function extractSpecialContext(&$entryList, &$holder)
	{
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			if ($entry->processTag('Alias') || $entry->processTag('AliasMatch') || $entry->processTag('ScriptAlias') || $entry->processTag('ScriptAliasMatch'))
			{
				$this->newAliasContext($entry, $holder);
			}
			elseif ($entry->processTag('Redirect') || $entry->processTag('RedirectMatch') || $entry->processTag('RedirectTemp') || $entry->processTag('RedirectPermanent'))
			{
				$this->newRedirectContext($entry, $holder);
			}
		}
	}

	/**
	 * Compare $path and $basePath to see if it is usable as a prefix.
	 *
	 * @param string $path The URI path
	 * @param string $basePath The document root for another context.
	 * @param string $baseUri The URI associated with $basePath
	 * @return mixed false if not match, new URI for $path if matched.
	 */
	function checkURIfromPath($path, $basePath, $baseUri)
	{
		if ((strlen($basePath) == 0)
				|| (strncmp($path, $basePath, strlen($basePath)) !== 0))
			return false;

		if (strlen($basePath) == strlen($path))
			return $baseUri;

		$uri = $baseUri . '/' . substr($path, strlen($basePath));
		$uri = str_replace(array('///', '//'), '/', $uri);
		return $uri;
	}

	/**
	 * Given the path and docroot, create a uri context.
	 *
	 * @param string $path The URI path
	 * @param string $docRoot The document root for the context.
	 * @param array $contextLists The destination array to extract to.
	 * @return mixed false if not match, new URI for $path if matched.
	 */
	function locateContext($path, $docRoot, &$contextLists)
	{
		$u = array_keys($contextLists);
		// find matched one first
		foreach ($u as $uri)
		{
			if ($path === $contextLists [$uri] ['location'])
				return $uri;
		}

		$found = false;
		foreach ($u as $uri)
		{
			$ctx = &$contextLists [$uri];
			$newUri = $this->checkURIfromPath($path, $ctx ['location'], $uri);
			if ($newUri !== false)
			{
				$found = true;
				break;
			}
		}

		if (!$found)
		{
			$newUri = $this->checkURIfromPath($path, $docRoot, '/');
			if ($newUri !== false)
			{
				// Can use this URI, create new context.
				$found = true;
				$ctx = array(
					'type' => 'NULL',
				);
			}
		}

		if ($found)
		{
			if (!isset($contextLists[$newUri]))
			{
				$ctx ['uri'] = $newUri;
				$ctx ['location'] = $path;
				$contextLists [$newUri] = $ctx;
			}
			return $newUri;
		}

		return false;
	}

	function containOption($val, $checkTag)
	{
		$pos = strpos($val, $checkTag);
		if ($pos !== false)
		{
			if ($pos > 0 && $val [$pos - 1] === '-')
				return 0;
			else
				return 1;
		}
		return - 1;
	}

	/**
	 * Process the allow override configuration.
	 *
	 * @param string $val The value string.
	 * @param array $ctx The context configuration.
	 */
	function extractCtxAllowOverride($val, &$ctx)
	{
		$res = $this->containOption($val, 'All');
		if ($res == 1)
			$allow = 31;
		else
			$allow = 0;
		$res = $this->containOption($val, 'Limit');
		if ($res == 1)
			$allow |= 1;
		elseif ($res == 0)
			$allow &= (~ 1);

		$res = $this->containOption($val, 'AuthConfig');
		if ($res == 1)
			$allow |= 2;
		elseif ($res == 0)
			$allow &= (~ 2);
		$res = $this->containOption($val, 'FileInfo');
		if ($res == 1)
			$allow |= 4;
		elseif ($res == 0)
			$allow &= (~ 4);
		$res = $this->containOption($val, 'Indexes');
		if ($res == 1)
			$allow |= 8;
		elseif ($res == 0)
			$allow &= (~ 8);
		$res = $this->containOption($val, 'Options');
		if ($res == 1)
			$allow |= 16;
		elseif ($res == 0)
			$allow &= (~ 16);
		$ctx ['allowOverride'] = $allow;
	}

	/**
	 * Extract access control settings.
	 *
	 * @param string|NULL $order The order configuration.
	 * @param string|NULL $allow The access allowed configuration.
	 * @param string|NULL $deny The access denied configuration.
	 * @param array $holder The array to put the configurations into.
	 */
	function extractAccessControl($order, $allow, $deny, &$holder)
	{
		if (!isset($order))
			return;
		if (isset($allow))
		{
			$allow = trim(substr($allow, 5)); // remove from
			$allow = str_replace(array('all', 'All', 'ALL'), '*', $allow);
			$holder ['accessControl_allow'] = $allow;
		}
		if (isset($deny))
		{
			$deny = trim(substr($deny, 5));
			$deny = str_replace(array('all', 'All', 'ALL'), '*', $deny);
			$holder ['accessControl_deny'] = $deny;
		}
	}

	/**
	 * Extract the authentication configurations for the context.
	 *
	 * @param array $entryList Configuration entries for the context.
	 * @param array $ctx The configuration array for the context.
	 * @param array $realms Current list of authentication realms for the VHost.
	 */
	function extractCtxAuth(&$entryList, &$ctx, &$realms)
	{
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			if ($entry->processTag('AuthName'))
			{
				$ctx ['authName'] = $entry->_val;
			}
			elseif ($entry->processTag('Require'))
			{
				$ctx ['required'] = $entry->_val;
			}
			elseif ($entry->processTag('AuthUserFile'))
			{
				$au = MigrUtil::getAbsPath($this->_apacheServerRoot, $entry->_val);
			}
			elseif ($entry->processTag('AuthGroupFile'))
			{
				$ag = MigrUtil::getAbsPath($this->_apacheServerRoot, $entry->_val);
			}
		}
		if (isset($au))
		{ //check realm
			$key = $au;
			if (isset($ag)) {
				$key .= $ag;
			}
			if (!isset($realms[$key]))
			{
				$realm = array();
				$realm ['userDB_location'] = $au;
				if (isset($ag))
					$realm ['groupDB_location'] = $ag;
				if (isset($ctx['authName']))
				{
					$realm_name = $ctx ['authName'];
					if (isset($realms ['names'] [$realm_name]))
						$realm_name .= '_' . count($realms ['names']);
				}
				else
					$realm_name = 'Realm_' . count($realms ['names']);
				$realm ['name'] = $realm_name;
				$realm ['type'] = 'file';
				$realms ['names'] [$realm_name] = $realm_name;

				$realms [$key] = $realm;
			}
		}
	}

	/**
	 * Process the context configuration for a normal context
	 *
	 * Apache uses Directory for context conf.
	 *
	 * @param Block $block The directory configuration block.
	 * @param array $ctx The configuration array for the context.
	 * @param array $realms Current list of authentication realms for the VHost.
	 */
	function extractDirectory(&$block, &$ctx, &$realms)
	{
		$entryList = &$block->_configEntries;
		$needAuth = false;
		$allow = null;
		$deny = null;
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			if ($entry->processTag('Options'))
			{
				$res = $this->containOption($entry->_val, 'Indexes');
				if ($res != - 1)
					$ctx ['autoIndex'] = $res;
			}
			elseif ($entry->processTag('AllowOverride'))
			{
				$this->extractCtxAllowOverride($entry->_val, $ctx);
			}
			elseif ($entry->processTag('AuthType'))
			{
				if ($entry->_val == 'Basic')
					$needAuth = true;
				else
					$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Not support Digest');
			}
			elseif ($entry->processTag('Order'))
			{
				$order = $entry->_val;
			}
			elseif ($entry->processTag('Allow'))
			{
				$allow = $entry->_val;
			}
			elseif ($entry->processTag('Deny'))
			{
				$deny = $entry->_val;
			}
		}
		if ($needAuth)
			$this->extractCtxAuth($entryList, $ctx, $realms);
		if (isset($order))
			$this->extractAccessControl($order, $allow, $deny, $ctx);
		$this->extractCtxRewrite($entryList, $ctx);
	}

	/**
	 * Extract the context configurations from the given Block.
	 *
	 * @param Block $block The virtual host block.
	 * @param array $vh The Apache VHost
	 */
	function extractContext(&$block, &$vh)
	{
		$contextLists = array();
		$realms = array();
		$this->extractSpecialContext($block->_configEntries, $contextLists);

		$dirBlocks = &$block->_childrenBlocks [Block::BT_DIR];

		if (isset($dirBlocks) && count($dirBlocks) > 0)
		{
			$dirLoc = array_keys($dirBlocks);
			foreach ($dirLoc as $loc)
			{
				if ($loc [0] === '~')
					$dirBlocks [$loc]->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Does not support regular expression Directory');
				else
				{
					$dirBlocks [$loc]->setStatus(MigrUtil::STATUS_PROCESSED, 'OK');
					$path = trim($dirBlocks [$loc]->_name, '"*\'');
					$uri = false;
					if (isset($vh ['general'] ['docRoot'])) {
						$uri = $this->locateContext($path, $vh ['general'] ['docRoot'], $contextLists);
					}
					if ($uri === false)
					{
						$dirBlocks [$loc]->setStatus(MigrUtil::STATUS_IGNORED, 'Ingores directory, not accessible!');
					}
					else
					{
						$ctx = &$contextLists [$uri];
						$this->extractDirectory($dirBlocks [$loc], $ctx, $realms);
					}
				}
			}
		}

		if (count($contextLists) > 0) {
			$vh ['context'] = array();
		}

		foreach ($contextLists as $ctx)
		{
			if ($ctx ['type'] == 'NULL')
				$ctx ['allowBrowse'] = 1;
			if (!isset($vh ['context'] [$ctx ['type']])) {
				$vh ['context'] [$ctx ['type']] = array();
			}
			$vh ['context'] [$ctx ['type']] [$ctx ['uri']] = $ctx;
		}
		if (count($realms) > 0)
		{
			if (!isset($vh ['security'])) {
				$vh ['security'] = array();
			}
			if (!isset($vh ['security'] ['realm'])) {
				$vh ['security'] ['realm'] = array();
			}
			foreach ($realms as $realm)
			{
				$vh ['security'] ['realm'] [$realm ['name']] = $realm;
			}
		}
	}

	/**
	 * Extract the VHost Rewrite configuration.
	 *
	 * @param array $entryList A VHost's configuration.
	 * @param array $holder The destination array to extract to.
	 */
	function extractVhRewrite(&$entryList, &$holder)
	{
		$rewrite = array();
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			elseif ($entry->processTag('RewriteEngine'))
			{
				if (!isset($rewrite ['ctrl'])) {
					$rewrite ['ctrl'] = array();
				}
				$rewrite ['ctrl'] ['enable'] = MigrUtil::getBool($entry->_val);
			}
			elseif ($entry->processTag('RewriteOptions'))
			{
				$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Not support at virtual host level');
			}
			elseif ($entry->processTag('RewriteLog'))
			{
				$entry->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, 'Not support, will write to error log.');
			}
			elseif ($entry->processTag('RewriteLogLevel'))
			{
				if (!isset($rewrite ['ctrl'])) {
					$rewrite ['ctrl'] = array();
				}
				$rewrite ['ctrl'] ['logLevel'] = $entry->_val;
			}
			elseif ($entry->processTag('RewriteLock'))
			{
				$entry->setStatus(MigrUtil::STATUS_IGNORED, '');
			}
			elseif ($entry->processTag('RewriteMap'))
			{
				$t = MigrUtil::splitToList($entry->_val);
				$map = array(
					'name' => $t [0],
					'location' => $t [1],
				);
				if (!isset($rewrite ['map'])) {
					$rewrite ['map'] = array();
				}
				$rewrite ['map'] [$map ['name']] = $map;
			}
			elseif ($entry->processTag('RewriteCond') || $entry->processTag('RewriteRule'))
			{
				if (!isset($rewrite ['rules'])) {
					$rewrite ['rules'] = '';
				}
				$rewrite ['rules'] .= $entry->_line . "\n";
			}
		}
		if (count($rewrite) > 0)
			$holder ['rewrite'] = $rewrite;
	}

	/**
	 * Extract the rewrite configurations for the context.
	 *
	 * @param array $entryList Configuration entries for the context.
	 * @param array $ctx The configuration array for the context.
	 */
	function extractCtxRewrite(&$entryList, &$ctx)
	{
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			elseif ($entry->processTag('RewriteEngine'))
			{
				$ctx ['rewrite_enable'] = MigrUtil::getBool($entry->_val);
			}
			elseif ($entry->processTag('RewriteOptions'))
			{
				if (strpos($entry->_val, 'inherit') !== false)
					$ctx ['rewrite_inherit'] = 1;
			}
			elseif ($entry->processTag('RewriteBase'))
			{
				$ctx ['rewrite_base'] = $entry->_val;
			}
			elseif ($entry->processTag('RewriteCond') || $entry->processTag('RewriteRule'))
			{
				$ctx ['rewrite_rules'] = $entry->_line;
			}
		}
	}

	/**
	 * Extract the Apache server security configurations to merge with LSWS conf.
	 *
	 * @param array $entryList The list of Entry objects to check against.
	 * @param array $s The destination array to extract to.
	 */
	function extractServSecurity(&$entryList, &$s)
	{
		$g = array();
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;
			elseif ($entry->processTag('Options'))
			{
				$res = $this->containOption($entry->_val, 'FollowSymLinks');
				if ($res != - 1)
					$s ['fileAccessControl_followSymbolLink'] = $res;
				$res = $this->containOption($entry->_val, 'SymLinksIfOwnerMatch');
				if ($res == 1)
					$s ['fileAccessControl_followSymbolLink'] = 2;
			}
		}
	}

	/**
	 * Extract server general settings from entry list to merge with LSWS conf.
	 *
	 * @param array $entryList Array of Entry objects.
	 */
	function extractServGeneral(&$entryList)
	{
		$byType = array();
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;

			if ($entry->processTag('ServerName'))
			{
				$this->_serv ['general'] ['serverName'] = $entry->_val;
			}
			elseif ($entry->processTag('ServerAdmin'))
			{
				$this->_serv ['general'] ['adminEmails'] = $entry->_val;
			}
			elseif ($entry->processTag('DirectoryIndex'))
			{
				$this->_serv ['general'] ['indexFiles'] = MigrUtil::getCommaList($entry->_val);
			}
			else
				$this->extractExpires($entry, $this->_serv ['general'], $byType);
		}
		if (!empty($byType))
			$this->_serv ['general'] ['expiresByType'] = implode(',', $byType);

		// $this->extractLogging( $entryList, $this->_serv['general'] );
	}

	/**
	 * Merge IP:Port listener configurations to LSWS conf.
	 *
	 * @param array $nameVHosts 2D array to indicate port and IP pair.
	 * @param Entry $entry The associated Entry for the configurations.
	 */
	private function appendListenerEntry($nameVHosts, &$entry)
	{
		$matches = array();
		if (preg_match("/^(\d+)$/", $entry->_val, $matches))
		{
			$ip = '*';
			$port = $matches [1];
			$listeners [] = new VHostConf('*', $matches [1]);
		}
		elseif (preg_match("/^([\d\.]+):(\d+)$/", $entry->_val, $matches))
		{
			$ip = (($matches [1] == '0.0.0.0') ? '*' : $matches [1]);
			$port = $matches [2];
		}
		else
		{
			$entry->setStatus(MigrUtil::STATUS_ERROR, 'Only accept IP Address:Port');
			return;
		}
		$addr = "$ip:$port";
		$listeners = &$this->_serv ['listeners'];
		if (isset($listeners [$addr]))
		{
			$entry->setStatus(MigrUtil::STATUS_ERROR, 'Address:Port already set.');
			return;
		}
		$listener = new VHostConf($ip, $port);
		$listeners [$addr] = &$listener;

		// TODO: previous logic did not approve non-wild name vhost IP with
		// wild listener IP. Should this be approved?
		if ((isset($nameVHosts[$port]))
			&& (is_int($nameVHosts[$port])
				|| (isset($nameVHosts[$port][$ip]))))
		{
			$listener->setIsNameVH();
		}
	}

	/**
	 * Merge Port listener configurations to LSWS Conf.
	 *
	 * This is for configurations with no IP set.
	 *
	 * @param array $nameVHosts 2D array to indicate port and IP pair.
	 * @param array $portEntries The associated Entry for the configurations.
	 */
	private function appendPortEntries($nameVHosts, &$portEntries)
	{
		$listeners = &$this->_serv ['listeners'];
		if (!empty($listeners))
		{
			foreach ($portEntries as &$port)
			{
				$port->setStatus(MigrUtil::STATUS_IGNORED,
						'Ignore due to Listen directive');
				return;
			}
		}

		foreach ($portEntries as &$port)
		{
			if (!ctype_digit($port->_val) // will also return false for neg numbers.
					|| (intval($port->_val) > 65535))
			{
				$port->setStatus(MigrUtil::STATUS_ERROR, 'Parse Error');
			}
			else if (isset($listeners['*:' . $port->_val]))
			{
				$port->setStatus(MigrUtil::STATUS_ERROR, 'Port already added');
			}
			else
			{
				$listener = new VHostConf('*', $port->_val);
				$listeners ['*:' . $port->_val] = &$listener;

				// TODO: previous logic did not approve non-wild name vhost IP
				// with wild listener IP. Should this be approved?

				if ((isset($nameVHosts[$port->_val]))
					&& (is_int($nameVHosts[$port->_val])))
				{
					$listener->setIsNameVH();
				}
			}
		}
	}

	/**
	 * Extract server listener list from entry list to merge with LSWS conf.
	 *
	 * @param array $entryList Array of Entry objects.
	 */
	function extractListener(&$entryList)
	{
		$listenerEntries = array();
		$portEntries = array();
		$nameVHosts = array();
		$this->_serv ['listeners'] = array();
		for ($i = 0; $i < count($entryList); ++$i)
		{
			$entry = &$entryList [$i];
			if ($entry->getStatus() > MigrUtil::STATUS_UNKNOWN)
				continue;

			if ($entry->processTag('Listen'))
			{
				$listenerEntries[] = &$entry;
				continue;
			}
			elseif ($entry->processTag('Port'))
			{
				$portEntries[] = &$entry;
				continue;
			}
			else if (!$entry->processTag('NameVirtualHost'))
				continue;
			$matches = array();
			if (preg_match("/^(\d+)$/", $entry->_val, $matches))
			{
				$port = $matches [1];
				if (!isset($nameVHosts [$port]))
					$nameVHosts [$port] = 1;
				else if (is_array($nameVHosts [$port]))
				{
					unset($nameVHosts [$port]);
					$nameVHosts [$port] = 1;
				}
			}
			elseif (preg_match("/^([\d\.]+):(\d+)$/", $entry->_val, $matches))
			{
				$ip = $matches [1];
				$port = $matches [2];
				if (!isset($nameVHosts [$port]))
				{
					$nameVHosts [$port] = array();
					$nameVHosts [$port] [$ip] = 1;
				}
				else if (is_array($nameVHosts [$port]))
					$nameVHosts [$port] [$ip] = 1;
				// if already a wildcard, ignore.
			}
			// TODO: are * ports allowed?
			else
				continue;
		}

		if (!empty($listenerEntries))
		{
			foreach ($listenerEntries as &$entry)
			{
				$this->appendListenerEntry($nameVHosts, $entry);
			}
		}

		if (!empty($portEntries))
		{
			$this->appendPortEntries($nameVHosts, $portEntries);
		}
	}

}
