<?php

/********************************************
* LiteSpeed AP2ZConf project
* @Author:   LiteSpeed Technologies, Inc. (http://www.litespeedtech.com)
* @Copyright: (c) 2017
*********************************************/

require_once('XmlTreeBuilder.php');
require_once('PathTool.php');
require_once('LSWSConf.php');
require_once('MigrUtil.php');
require_once('VHostConf.php');
require_once('ParsedConf.php');
require_once('FileLoader.php');
require_once('LSWSExtractor.php');


class Migrator
{

	var $_rawData;
	var $_extractor;
	var $_sslConfs = array();
	var $_apacheServerRoot;
	var $_ipMap;
	// input var
	var $_lswsServerRoot;
	var $zconfMsgDir;
	var $apache_conf_path;
	var $exclude_domains;
	var $server_cluster;
	var $max_conns;

	/**
	 * Initializes the Migrator class.
	 *
	 * @param string $apacheConfPath Absolute path to the main apache configurations file.
	 * @param string $lswsServerRoot Absolute path to LSWS server root.
	 * @param string $zconfDir Directory to store the output zconf messages.
	 */
	function init($apacheConfPath, $lswsServerRoot, $zconfDir = null)
	{
		$this->_lswsServerRoot = $lswsServerRoot;
		$this->apache_conf_path = $apacheConfPath;
		if ($zconfDir)
			$this->zconfMsgDir = realpath($zconfDir);
		else
			$this->zconfMsgDir = realpath($lswsServerRoot) . '/conf';
	}

	function setIpMap($ipMap)
	{
		$this->_ipMap = $ipMap;
	}

	function setExcludeDomains($excludeDomains)
	{
		$this->exclude_domains = $excludeDomains;
	}

	function setServerCluster($clusterips)
	{
		$this->server_cluster = $clusterips;
	}

	function setDefaultMaxConns($maxconns)
	{
		$this->max_conns = $maxconns;
	}

	function renameOldFile(&$filename)
	{
		$bak = '.old';
		$i = 1;

		if (!file_exists($filename))
		{
			return true;
		}

		if (file_exists($filename . $bak))
		{
			$bak = sprintf(".old_%02d", $i);
			while (file_exists($filename . $bak))
			{
				$i++;
				$bak = sprintf(".old_%02d", $i);
			}
		}

		if (($i <= 10) || !class_exists('ZipArchive'))
		{
			$ret = copy($filename, $filename . $bak);
			return $ret;
		}

		$zip = new ZipArchive;
		$dir = dirname($filename);
		$arr = scandir($dir);
		if (false === $arr) {
			return false;
		}
		$pattern = '/' . basename($filename) . '\.old_[0-9]+/';
		echo('pattern is ' . $pattern);
		$parsed = preg_grep($pattern, $arr);

		if (empty($parsed))
		{
			return false;
		}

		$res = $zip->open($filename . '_bak.zip', ZipArchive::CREATE | ZipArchive::OVERWRITE);
		if ( $res !== true )
		{
			echo('Warning: Failed to archive backups for ' . $filename);
			$ret = copy($filename, $filename . $bak);
			return $ret;
		}

		foreach ($parsed as $key => $val)
		{
			$parsed[$key] = $dir . '/' . $val;
			if (!$zip->addFile($parsed[$key], $val))
			{
				echo('Warning: Failed to archive backup file ' . $val);
				$zip->close();
				$ret = copy($filename, $filename . $bak);
				return $ret;
			}
		}

		$ret = $zip->close();
		if (!$ret)
		{
			echo('Warning: Failed to close archive.');
			return $ret;
		}
		$bak = '.old_01';

		foreach ($parsed as $delFile)
		{
			unlink($delFile);
		}

		$ret = copy($filename, $filename . $bak);
		return $ret;
	}

	/**
	 * Load the local server configurations.
	 *
	 * The Apache configurations are loaded first, then LSWS configurations.
	 *
	 * The data is stored as Blocks.
	 */
	function loadApacheConf()
	{
		$fileLoader = new FileLoader ();
		$this->_rawData = &$fileLoader->load($this->apache_conf_path);
		$this->_apacheServerRoot = $fileLoader->_apacheServerRoot;
		$fileLoader = NULL;

		$this->_extractor = new Extractor($this->_apacheServerRoot, $this->_lswsServerRoot);
		$this->_extractor->extract($this->_rawData);

	}

	/**
	 * Create LSWS configuration files based on the Apache conf files.
	 */
	function createLSWSConf()
	{
		$writer = new LSWSConf();
		$filename = $this->_lswsServerRoot . "/conf/httpd_config.xml";
		if (!$this->renameOldFile($filename))
		{
			echo 'ERROR: failed to back up old lsws config file.';
			die();
		}
		$writer->saveServerFile($this->_extractor->_serv, $filename);
		$vhnames = array_keys($this->_extractor->_vhosts);
		foreach ($vhnames as $vhname)
		{
			$filename = $this->_lswsServerRoot . "/conf/$vhname.xml";
			if (!$this->renameOldFile($filename))
			{
				echo 'ERROR: failed to back up old lsws vhost config file.';
				die();
			}
			$writer->saveVhFile($vhname, $this->_extractor->_vhosts[$vhname], $filename);
		}

	}

	/**
	 * Check the configuration block for the list of IP and port pairs.
	 *
	 * @param string $ipList A string of IP port pairs.
	 * @return mixed NULL if no valid IPs, space delimited list of IP:port pairs otherwise.
	 */
	private function parseVHostIPList($ipList)
	{
		$unallowedV6 = array(
			'::', // wildcard
			'::1' // localhost
		);
		$unallowedV4 = array(
			'0.0.0.0', // wildcard
			'127.0.0.1' // localhost
		);
		$key = array();
		$vhost_list = explode(" ", $ipList);
		foreach ($vhost_list as $vhost)
		{
			$ip = NULL;
			$port = NULL;
			$isV6 = ($vhost[0] == '[');

			if ($isV6)
			{
				if (!defined('AF_INET6'))
					continue; // ipv6 not compiled
				$scanFormat = '[%[^]]]:%d';
				$unallowed = &$unallowedV6;
			}
			else
			{
				$scanFormat = '%[^:]:%d';
				$unallowed = &$unallowedV4;
			}

			$numSet = sscanf($vhost, $scanFormat, $ip, $port);
			if ($numSet == 0)
				continue; // did not have an IP and port.
			elseif ($numSet == 1)
				$port = '';

			// REMARK: inet_pton will return a warning if the IP is not properly
			// formatted. I would prefer to use filter_var first to avoid
			// needing to do this error handling, but that requires the
			// filter module, which is not guaranteed to be loaded.
			set_error_handler(
					function($errno, $errstr)
					{
//						error_log('DEBUG: error: ' . $errno . ': ' . $errstr);
					});
			$ipBin = inet_pton($ip);
			restore_error_handler();

			if (false === $ipBin)
				continue; // not a valid IP

			// Make IP consistent.
			$ip = inet_ntop($ipBin);

			if (in_array($ip, $unallowed))
				continue;

			$key[] = ($isV6 ? "[$ip]:$port" : "$ip:$port");
		}

		return (empty($key) ? NULL : implode(' ', $key));
	}

	/**
	 * Add the SSL configurations for this VHost associated with the given
	 * VHost name and domain list.
	 *
	 * @param string $name The Apache VHost name - typically ip:port
	 * @param array $domainList The domains added
	 * @param array $sslConf the ssl configurations for the vhost.
	 */
	private function appendSslConf($name, $domainList, &$sslConf)
	{
		$keyFile = (isset($sslConf['sslcertificatekeyfile'])
				? $sslConf['sslcertificatekeyfile'] : null);
		$certFile = (isset($sslConf['sslcertificatefile'])
				? $sslConf['sslcertificatefile'] : null);
		$cabundleFile = (isset($sslConf['sslcertificatechainfile'])
				? $sslConf['sslcertificatechainfile'] : null);

		if (is_null($keyFile) && is_null($certFile))
			return;
		else if (is_null($keyFile))
			$keyFile = &$certFile;

		$keyContents = file_get_contents($keyFile);

		if ($keyContents === false)
			return;

		$matches = array();

		if (!preg_match(
				'/(-{1,5}BEGIN (?:RSA )?PRIVATE KEY-{1,5}[^-]*-{1,5}END (?:RSA )?PRIVATE KEY-{1,5})/m',
				$keyContents, $matches)) // If it returns 0 or false, both are bad.
			return;

		$key = $matches[1];

		if ($certFile === $keyFile)
			$certContents = &$keyContents;
		else
			$certContents = file_get_contents($certFile);

		if (false === $certContents) {
			return;
		}

		$certmatches = array();
		$certMatchCount = preg_match_all(
			'/(-{1,5}BEGIN CERTIFICATE-{1,5}[^-]*-{1,5}END CERTIFICATE-{1,5})/m',
			$certContents, $certmatches);
		if (!$certMatchCount) // If it returns 0 or false, both are bad.
            return;
        $certmatches = $certmatches[1];

		$cert = $certmatches[0];

        if (is_null($cabundleFile))
            $cabundleFile = $certFile;

		if ($cabundleFile)
		{
			if ($cabundleFile === $certFile)
			{
				if ($certMatchCount > 1)
					$cabmatches = array_slice($certmatches, 1);
			}
			else
			{
				$cabmatches = array();
				if ($cabundleFile === $keyFile)
					$cabContents = &$keyContents;
				else
					$cabContents = file_get_contents($cabundleFile);

				if (false === $cabContents) {
					return;
				}

				$cabMatchCount = preg_match_all(
					'/(-{1,5}BEGIN CERTIFICATE-{1,5}[^-]*-{1,5}END CERTIFICATE-{1,5})/m',
					$cabContents, $cabmatches);
				if (!$cabMatchCount)
					$cabmatches = null;
                else
                {
                    $cabmatches = $cabmatches[1];
                    array_shift($cabmatches);
                }
			}
			if (isset($cabmatches) && !empty($cabmatches))
				$cabundle = implode("\n", $cabmatches);
		}

		$confStr = "\n\t\t{\n\t\t\"domain_list\" :\n\t\t\t[ "
			. implode(',', $domainList)
			. "],\n"
			. "\t\t\"key\" : \"" . $key . "\",\n"
			. "\t\t\"cert\" : \"" . $cert . "\""
			. (isset($cabundle) ? ",\n\t\t\"ca_bundle\" : \"" . $cabundle . "\"" : '')
			. "\n\t\t}";
		$this->_sslConfs[$name] = $confStr;
	}

	/**
	 * Extract an Apache VHost configuration.
	 *
	 * @param Block $vhostEntry The VHost to extract from.
	 * @param string $apName The Apache VHost name - uses the ServerName parameter
	 * @param array &$domainList The domain list to append to.
	 */
	private function getDomains($vhostEntry, $apName, &$domainList)
	{
		$vhostList = &$this->_extractor->_vhosts;
		$vhost = null;
		$cnt = 0;
		foreach ($vhostEntry->_configEntries as $directive)
		{
			if ('ServerName' != $directive->_directive)
			{
				continue;
			}
			if ((isset($vhostList[$directive->_val])) &&
					($vhostList[$directive->_val]['apName'] == $apName))
			{
				$vhost = &$vhostList[$directive->_val];
				break;
			}
			$suffix = 1;
			while (isset($vhostList[$directive->_val . $suffix]))
			{
				if ($vhostList[$directive->_val . $suffix]['apName']
						== $apName)
				{
					$vhost = &$vhostList[$directive->_val . $suffix];
					break;
				}
				++$suffix;
			}
			break;
		}
		$default_cluster = (isset($this->server_cluster) ? $this->server_cluster : 'default');
		$cluster = (isset($vhost['cluster']) ? $vhost['cluster'] : $default_cluster); // Sets to template if set, else default.
		$template = (isset($vhost['template']) ? $vhost['template'] : 'default'); // Sets to template if set, else default.
		if (is_null($vhost))
			return;
		foreach ($vhost['domains'] as $domain)
		{
			if (($domain == "*") // no wildcard domains
				|| (substr($domain, -1) === '*') // no trailing wildcards
				|| (preg_match('/[^[\]:\d\.]/', $domain) == 0) // no IPs
				|| (isset($this->exclude_domains[$domain])))
			{
				continue;
			}
			if (!isset($domainList[$cluster])) {
				$domainList[$cluster] = array();
			}
			if (!isset($domainList[$cluster][$template])) {
				$domainList[$cluster][$template] = array();
			}
			$domainList[$cluster][$template][] = "\"$domain\"";
			++$cnt;
		}

		if (($cnt <= 0) || (isset($this->_sslConfs[$vhost['vhTop']['name']]))
				|| !(isset($vhost['ssl'])))
		{
			return;
		}

		$this->appendSslConf($vhost['vhTop']['name'],
				array_slice($domainList[$cluster][$template], -$cnt), $vhost['ssl']);
	}

	private function getVhostIp($vhost_ip)
	{
		if (!empty($this->_ipMap) && isset($this->_ipMap[$vhost_ip]))
		{
			$vhost_ip = $this->_ipMap[$vhost_ip];
		}
		return $vhost_ip;
	}

	/**
	 * Create a configuration that lists all domains associated with a ip:port pair.
	 *
	 * @param array $vhost_list The list of vhosts
	 * @param int|NULL $isSecure Returns whether the VHostConf is a secure or insecure listener.
	 * @return mixed Null on failure or empty conf, compiled vhost configuration on success.
	 */
	private function buildConfList($vhost_list, &$isSecure)
	{
		$listenerList = $this->_extractor->_serv['listeners'];
		$port_list = array();
		$ip_list_entries = array();
		$isSecure = NULL;
		foreach ($vhost_list as $vhost)
		{
			$entry = explode(';', $vhost);
			$portStart = strrpos($entry[0], ':');
			if (false === $portStart)
				continue;

			$vhost_ip = substr($entry[0], 0, $portStart);
			$vhost_port = substr($entry[0], $portStart + 1);
			$listenerIdx = $entry[0];
			if (!isset($listenerList[$listenerIdx]))
			{
				// TODO: Is this only valid for ipv4?
				$listenerIdx = "*:$vhost_port";
			}

			if (isset($listenerList[$listenerIdx]))
			{
				if (is_null($isSecure))
				{
					$isSecure = $listenerList[$listenerIdx]->isSecure();
				}
				else if ($isSecure != $listenerList[$listenerIdx]->isSecure())
				{
					error_log('ZCONF: ERROR - vhost listener list secure settings'
							. 'do not match!');
					return null;
				}

			}
			else
			{
				error_log('ZCONF: Warning! listener not found in configuration.');
			}

			if (false !== $vhost_port && !empty($vhost_port))
			{
				$port_str = ", \"port_list\" : [$vhost_port]";
				if (!in_array($vhost_port, $port_list))
				{
					$port_list[] = $vhost_port;
				}
			}
			else
			{
				$port_str = '';
			}
			$vhost_ip = $this->getVhostIp($vhost_ip);
			$max_conns = $this->max_conns;

			if (count($entry) > 1) {
				foreach ($entry as $attr) {
					if (strlen($attr) > 9 && strncmp($attr, 'max_conn=', 9) == 0) {
						$max_conns = intval(substr($attr, 9));
						break;
					}
				}
			}

			$max_conn_str = '';
			if ($this->max_conns != 0) {
				$max_conn_str = ', "max_conn" : ' . $max_conns;
			}
			$ip_list_entries[] = "\t\t\t\t\t{ \"ip\": \"$vhost_ip\"$port_str$max_conn_str }";
		}
		if (empty($ip_list_entries))
		{
			return null;
		}
		return sprintf(
				"\t\t\t\"lb_port_list\" : [%1\$s],\n"
				. "\t\t\t\"dport\" : %1\$s,\n"
				. "\t\t\t\"be_ssl\" : %2\$s,\n"
				. "\t\t\t\"ip_list\" :\n\t\t\t\t[\n%3\$s"
				. "\n\t\t\t\t]\n\t\t\t}\n\t\t]\n\t\t}\n",
				implode(",", $port_list),
				(($isSecure) ? "true" : "false"),
				implode(",\n", $ip_list_entries)
				);
	}

	/**
	 * Create the SSL message file.
	 */
	private function createSslMsg()
	{
		$confStr = "conf=\n{\n\"ssl_list\" :\n\t["
				. implode(',', $this->_sslConfs)
				. "\n\t]\n}";
		$sslFile = $this->zconfMsgDir . "/sslconf.msg";
		if (!$this->renameOldFile($sslFile))
		{
			echo 'ERROR: failed to back up old ssl config file.';
			return false;
		}
		$fd = fopen($sslFile, 'w');
		if (!$fd)
		{
			return false;
		}
		fwrite($fd, $confStr);
		fclose($fd);
		return true;
	}

	/**
	 * Create the SSL release message. for the given domains.
	 *
	 * @param array $domList The list of domains to release.
	 * @param string $fileName The file to update.
	 */
	private function genSslReleaseFile($domList, $fileName)
	{
		$conf = "conf=\n{\n\"ssl_list\" :\n\t[\n\t\t{\n\t\t\t\"domain_list\" :[ "
			. implode(',', $domList)
			. " ]\n\t\t}\n\t]\n}";

		$fd = fopen($fileName, 'w');
		if (!$fd)
		{
			return;
		}
		fwrite($fd, $conf);
		fclose($fd);

	}

	/**
	 * Generate messages to force update SSL and SSL entries to remove.
	 *
	 * The force update SSL domains are the ones to be added.
	 *
	 * The SSL entries to remove are the entries that already existed, but no
	 * longer.
	 *
	 * @param array $curDoms The list of domains that will be updated.
	 * @param string $confStr The contents of the last sent configuration.
	 */
	private function createSslReleaseMsg(&$curDoms, $confStr)
	{
		$parsed = json_decode(substr($confStr, 6), true);
		$ssl_domains = null;
		if (is_null($parsed)) {
			return;
		}
		foreach ($parsed['vhost_list'] as $vhost)
		{
			if ( $vhost['conf_list'][0]['be_ssl'] )
			{
				$ssl_domains = $vhost['domain_list'];
				break;
			}
		}
		if (is_null($ssl_domains))
			return;

		$this->genSslReleaseFile($curDoms, $this->zconfMsgDir . "/forceupdatessl.msg");

		$cmpDoms = array();

		foreach ($ssl_domains as $domain)
			$cmpDoms[] = "\"$domain\"";

		$removedDoms = array_diff($cmpDoms, $curDoms);
		if (empty($removedDoms))
			return;

		$this->genSslReleaseFile($removedDoms, $this->zconfMsgDir . "/sslrelease.msg");
	}

	/**
	 * Generate ZConf Message files based on Apache configs.
	 */
	function createZConfMsg()
	{
		// SECTION: Start Config Setup
		$config_vhost_list = array(); // vhosts/main domains
		foreach ($this->_rawData->_childrenBlocks [Block::BT_VHOST] as $key => $virtualHost)
		{
			$parsedKey = $this->parseVHostIPList($key);
			if (is_null($parsedKey))
				continue;
			foreach ($virtualHost as $entry)
			{
				if (!isset($config_vhost_list[$parsedKey]))
					$config_vhost_list[$parsedKey] = array();
				$this->getDomains($entry, $key, $config_vhost_list[$parsedKey]);
			}
		}
		// End Config Setup
		// SECTION: Start of Config String Building
		$confStr = "conf=\n{\n\"vhost_list\" :\n\t[\n";
		$conf_ssls = array();
		$isSecure = null;
		$confs = array();
		// config_vhost_list holds listener ip:port as key, template list as val.
		foreach ($config_vhost_list as $key => $clusterlist)
		{
			$vhostIps = explode(' ', $key);
			$portStart = strrpos($vhostIps[0], ':');
			if ($portStart !== false)
				$vhost_port = substr($vhostIps[0], $portStart);
			else
				$vhost_port = '';

			foreach ($clusterlist as $cluster => $templatelist)
			{
				if ($cluster != 'default')
				{
					$wantKey = true;
					if (substr($cluster, -1) == '$')
					{
						$cluster = substr($cluster, 0, -1);
						$wantKey = false;
					}
					$iplist = explode(',', $cluster);
					if (!empty($vhost_port))
					{
						foreach ($iplist as $k => $ip)
						{
							$ip = trim($ip);
							$attrs = strpos($ip, ';');
							if ($attrs !== false) {
								$segment = substr($ip, 0, $attrs);
								$ip = trim($segment) . $vhost_port . substr($ip, $attrs);
							} else {
								$ip .= $vhost_port;
							}
							$iplist[$k] = $ip;
						}
					}
					if ($wantKey)
					{
						$iplist = array_merge($iplist, $vhostIps);
					}
				}
				else
					$iplist = $vhostIps;
				foreach ($templatelist as $template => $domainlist)
				{
					if (!isset($domainlist) || empty($domainlist))
					{
						continue;
					}
					$confStr_domain_list = "\t\t{";
					if ($template != 'default')
						$confStr_domain_list .= "\n\t\t\"template\" : \"$template\",";

					$confStr_domain_list .= "\n\t\t\"domain_list\" :\n\t\t\t[ "
						. implode(',', $domainlist)
						. "],\n\t\t\"conf_list\" : [\n\t\t\t{\n";
					if ($confStr_domain_list === false)
						return false;

					$confStr_conf_list = $this->buildConfList($iplist, $isSecure);
					if (is_null($confStr_conf_list))
						continue;

					if ($isSecure)
						$conf_ssls = array_unique(array_merge($conf_ssls, $domainlist));

					$confs[] = $confStr_domain_list . $confStr_conf_list;
				}
			}
		}
		$confStr .= implode(',', $confs);
		$confStr .= "\n\t]\n}";
		// End of Config String Building
		// Write Config String to a .msg file for zconf
		$confFile = $this->zconfMsgDir . "/confstr.msg";

		// Remove force SSL update file.
		if (file_exists($this->zconfMsgDir . "/forceupdatessl.msg"))
			unlink($this->zconfMsgDir . "/forceupdatessl.msg");
		// Remove no longer configured domains SSL release file.
		if (file_exists($this->zconfMsgDir . "/sslrelease.msg"))
			unlink($this->zconfMsgDir . "/sslrelease.msg");
		// Create force SSL update and SSL release files.
		if (file_exists($confFile) && !empty($conf_ssls))
		{
			$confFileContents = file_get_contents($confFile);
			if ($confFileContents === false) {
				echo 'ERROR: failed to read existing conf file.';
				return false;
			}
			$this->createSslReleaseMsg($conf_ssls, $confFileContents);
		}

		if (!$this->renameOldFile($confFile))
		{
			echo 'ERROR: failed to back up old zconf config file.';
			return false;
		}
		$fd = fopen($confFile, 'w');
		if (!$fd)
		{
			return false;
		}
		fwrite($fd, $confStr);
		fclose($fd);
		$confStr = null;
		// End of file write

		return $this->createSslMsg();
	}

}

