<?php

/********************************************
* LiteSpeed AP2ZConf project
* @Author:   LiteSpeed Technologies, Inc. (http://www.litespeedtech.com)
* @Copyright: (c) 2017
*********************************************/

require_once('MigrUtil.php');

# An entry takes a single configuration extracted from the Apache conf.
class Entry
{

	private static $IgnoredTags = array(
				'ServerType', 'LockFile', 'PidFile', 'ScoreBoardFile',
				'KeepAlive', 'MaxKeepAliveRequests', 'KeepAliveTimeout',
				'MinSpareServers', 'MaxSpareServers', 'StartServers',
				'MaxClients', 'MaxRequestsPerChild', 'BindAddress',
				'LoadModule', 'User', 'Group', '</IfDefine>'
	);

	private static $NotSupportedTags = array(
				'ServerPath', 'CookieLog', 'LogFormat'
	);

	var $_directive;
	var $_val;
	var $_status; // Unknown=0, Processed=1, Ignored=2, Error=3, Not Support=4
	var $_statusMsg;
	var $_line;
	var $_lineNum;
	var $_fileName;

	/**
	 * Create an Entry for the configuration line.
	 * 
	 * Sets the status based on if the tag is supported.
	 *
	 * @param string $line The Entry's full line.
	 * @param int $lineIndex The Entry's line number within the file.
	 * @param string $fileName The name of the file.
	 */
	public function __construct($line, $lineIndex, &$fileName)
	{
		$this->_line = $line;
		$this->_lineNum = $lineIndex;
		$this->_fileName = &$fileName;

		$pos = strpos($line, ' ');
		if (false === $pos) {
			$this->setStatus(MigrUtil::STATUS_ERROR, null);
			return;
		}
		$this->_directive = substr($line, 0, $pos);
		$this->_val = trim(substr($line, $pos), " \"\'");

		if (self::isTagIgnored($this->_directive))
			$this->setStatus(MigrUtil::STATUS_IGNORED, null);
		elseif (self::isTagNotSupported($this->_directive))
			$this->setStatus(MigrUtil::STATUS_NOT_SUPPORTED, null);
		else
			$this->setStatus(MigrUtil::STATUS_UNKNOWN, null);
	}

	function processTag($tag)
	{
		if (strcasecmp($this->_directive, $tag) == 0)
		{
			$this->setStatus(MigrUtil::STATUS_PROCESSED, null);
			return true;
		}
		return false;
	}

	/**
	 * Set the Entry's status and status message.
	 *
	 * @param int $status The Entry's new status.
	 * @param string|NULL $statusMsg The message to associate with the status.
	 */
	function setStatus($status, $statusMsg)
	{
		$this->_status = $status;
		$this->_statusMsg = $statusMsg;
	}

	function getStatus()
	{
		return $this->_status;
	}

	/**
	 * Check if the tag should be ignored.
	 * 
	 * Ignored tags may be relevant to ZConf in the future, but not now.
	 *
	 * @param string $tag The tag to check.
	 */
	static function isTagIgnored($tag)
	{
		return in_array($tag, self::$IgnoredTags);
	}

	/**
	 * Check if the tag is not supported.
	 * 
	 * Unsupported tags are not relevant to ZConf.
	 *
	 * @param string $tag The tag to check.
	 */
	static function isTagNotSupported($tag)
	{
		return in_array($tag, self::$NotSupportedTags);
	}
}

# A block holds a collection of configurations. It may hold entries or nested blocks.
class Block
{
	const BT_MODULE = 1;
	const BT_VHOST = 2;
	const BT_DIR = 3;
	const BT_FILES = 4;
	const BT_FILESMATCH = 5;
	const BT_LOCATION = 6;

	var $_type;
	var $_name;
	var $_depth;
	var $_allowDup;
	var $_configEntries = array();
	var $_childrenBlocks = array();
	var $_status; // Unknown=0, Processed=1, Ignored=2, Error=3, Not Support=4
	var $_statusMsg;

	/**
	 * Initializes the Migrator class.
	 *
	 * @param string $type Block Type used to distinguish and index Blocks.
	 * @param string $name Block Name used as a second level of indexing.
	 * @param int $depth The depth level of the current Block.
	 * @param bool $dupKey Indicate whether multiple blocks with the same 
	 *     Type and Name are permitted.
	 */
	public function __construct($type, $name, $depth, $dupKey)
	{

		switch (strtolower($type))
		{
		case 'ifmodule':
			$blockType = self::BT_MODULE;
			break;
		case 'virtualhost':
			$blockType = self::BT_VHOST;
			break;
		case 'directory':
			$blockType = self::BT_DIR;
			break;
		case 'files':
			$blockType = self::BT_FILES;
			break;
		case 'filesmatch':
			$blockType = self::BT_FILESMATCH;
			break;
		case 'location':
			$blockType = self::BT_LOCATION;
			break;
		default:
			$blockType = $type;
			break;
		}
		$this->_type = $blockType;
		$this->_name = $name;
		$this->_depth = $depth;
		$this->_allowDup = $dupKey;
		if ($depth > 100)
			exit("Failed to parse the file: reached maximum depth $depth!");
	}

	/**
	 * Append the Entry to the list of configuration options.
	 *
	 * @param Entry $entry The entry to append.
	 */
	function addAttribute(&$entry)
	{
		$this->_configEntries [] = $entry;
	}

	function setStatus($status, $statusMsg)
	{
		// Unknown=0, Processed=1, Ignored=2, Error=3, Not Support=4
		$this->_status = $status;
		$this->_statusMsg = $statusMsg;
	}

	function getEndTag()
	{
		if ($this->_depth == 0)
			return null;
		if (is_string($this->_type))
			return '</' . $this->_type . '>';

		switch ($this->_type)
		{
		case self::BT_MODULE:
			return '</IfModule>';
		case self::BT_VHOST:
			return '</VirtualHost>';
		case self::BT_DIR:
			return '</Directory>';
		case self::BT_FILES:
			return '</Files>';
		case self::BT_FILESMATCH:
			return '</FilesMatch>';
		case self::BT_LOCATION:
			return '</Location>';
		default:
			return null;
		}
	}

	/**
	 * Append $block to this Block's children blocks list.
	 * 
	 * If the child Block is a Module Block, its children are appended to this
	 * Block's children list, rather than appending the child Block itself.
	 *
	 * @param Block $block The child block to append.
	 */
	function addChildBlock(&$block)
	{
		if ($block->_type == Block::BT_MODULE)
		{ // move one level up
			if (count($block->_configEntries) > 0)
			{
				foreach ($block->_configEntries as $entry)
					$this->_configEntries [] = $entry;
			}
			if (count($block->_childrenBlocks) > 0)
			{
				$types = array_keys($block->_childrenBlocks);
				foreach ($types as $type)
				{
					$names = array_keys($block->_childrenBlocks [$type]);
					foreach ($names as $name)
					{
						if (is_array($block->_childrenBlocks[$type][$name]))
						{
							foreach ($block->_childrenBlocks [$type] [$name] as $child)
								$this->_childrenBlocks [$type] [$name] [] = $child;
						}
						else
						{
							$this->_childrenBlocks [$type] [$name] = $block->_childrenBlocks [$type] [$name];
						}
					}
				}
			}
		}
		else
		{
			if (!isset($this->_childrenBlocks [$block->_type])) {
				$this->_childrenBlocks [$block->_type] = array();
			}
			if ($block->_allowDup) {
				if (!isset($this->_childrenBlocks [$block->_type] [$block->_name])) {
					$this->_childrenBlocks [$block->_type] [$block->_name] = array();
				}
				$this->_childrenBlocks [$block->_type] [$block->_name] [] = $block;
			}
			else
				$this->_childrenBlocks [$block->_type] [$block->_name] = $block;
		}
	}

}
