<?php

/********************************************
* LiteSpeed AP2ZConf project
* @Author:   LiteSpeed Technologies, Inc. (http://www.litespeedtech.com)
* @Copyright: (c) 2017
*********************************************/

class XmlTreeBuilder
{

	var $parser;
	var $node_stack;

	/**
	 * Load and parse an XML file.
	 *
	 * @param string $filename The file to parse.
	 */
	function &parseFile($filename)
	{
		$root = NULL;
		if (!is_file($filename))
			return $root;
		$fd = fopen($filename, 'r');
		if (!$fd)
			return $root;

		$contents = fread($fd, filesize($filename));
		fclose($fd);

		if ($contents !== false) {
			$root = $this->parseString($contents);
		}
		return $root;
	}

	/**
	 * Take the XML string and build an associated array (multi-dimensional) of the configurations.
	 *
	 * This class holds the XML parser as well as the associated array.
	 * While the XML parser parses the string, the parser will call this class'
	 * callback functions at the start of an element, at the end of an element,
	 * and when character data needs to be handled.
	 * 
	 * @param string $xmlstring The string to parse.
	 * @return array Returns the fully parsed associated array for the XML string.
	 */
	function parseString(&$xmlstring)
	{
		$this->parser = xml_parser_create();
		xml_set_object($this->parser, $this);
		xml_parser_set_option($this->parser, XML_OPTION_CASE_FOLDING, false);
		xml_set_element_handler($this->parser, 'startElement', 'endElement');
		xml_set_character_data_handler($this->parser, 'characterData');

		$this->node_stack = array();
		$this->startElement(null, 'root', array());

		xml_parse($this->parser, $xmlstring);
		xml_parser_free($this->parser);

		$rnode = array_pop($this->node_stack);
		if (is_null($rnode)) {
			return NULL;
		}

		return($rnode['ELEMENTS'][0]);
	}

	/**
	 * Callback called at the start of a new element.
	 * 
	 * Appends a new node to the node stack with the given tag. The attributes
	 * are set as associated entries in the node array.
	 *
	 * @param resource|NULL $parser Reference to the XML parser.
	 * @param string $name The name of the new element.
	 * @param array $attrs The attributes of the new element.
	 */
	function startElement($parser, $name, $attrs)
	{
		$node = array();
		$node['TAG'] = $name;

		foreach ($attrs as $key => $value)
		{
			$node[$key] = $value;
		}

		$node['VALUE'] = '';
		$node['ELEMENTS'] = array();

		$this->node_stack[] = $node;
	}

	/**
	 * Callback called at the end of an element.
	 * 
	 * Remove the node from the end of the node stack and add it to the parent
	 * (previous node)'s elements list.
	 *
	 * @param resource|NULL $parser Reference to the XML parser.
	 * @param string $name The name of the completed element.
	 */
	function endElement($parser, $name)
	{
		$node = array_pop($this->node_stack);
		$node['VALUE'] = trim($node['VALUE']);

		$lastnode = count($this->node_stack);
		$this->node_stack[$lastnode - 1]['ELEMENTS'][] = $node;
	}

	/**
	 * Callback called when character data is processed.
	 * 
	 * Appends the new character data to the value of the node.
	 *
	 * @param resource|NULL $parser Reference to the XML parser.
	 * @param string $data The newly processed character data.
	 */
	function characterData($parser, $data)
	{
		$lastnode = count($this->node_stack);
		$this->node_stack[$lastnode - 1]['VALUE'] .= $data;
	}

}

?>