package zconflib::utils;

####################################################
# Copyright 2017 LiteSpeed Technologies
# http://www.litespeedtech.com
####################################################

use strict;
use warnings;

use Archive::Zip                    qw(:ERROR_CODES);
use CGI                             qw(param header);
use Encode                          qw(encode);
use File::Basename                  qw(fileparse);
use POSIX                           qw(strftime);
use Storable                        qw(retrieve store);

use Cpanel::JSON                    qw(Dump);

use Whostmgr::ACLS();

my $tpDir = '/usr/local/cpanel/3rdparty/webadc_zconf/';

my $pluginConfFile = $tpDir . 'confs/zconf.conf';
my $serverUpFile = $tpDir . 'confstr.msg';
my $sslForceUpdateFile = $tpDir . 'forceupdatessl.msg';
my $sslMsgFile = $tpDir . 'sslconf.msg';
my $sslReleaseMsgFile = $tpDir . 'sslrelease.msg';
my $logFileName = $tpDir . 'webadc_zconf.log';

my $logFH;

sub getThirdPartyDir {
    return $tpDir;
}

sub getPluginConfFile {
    return $pluginConfFile;
}

sub getServerUpFile {
    return $serverUpFile;
}

sub getSslForceUpdateFile {
    return $sslForceUpdateFile;
}

sub getSslMsgFile {
    return $sslMsgFile;
}

sub getSslReleaseMsgFile {
    return $sslReleaseMsgFile;
}

sub formatMsg {
    my ($unformatted) = @_;
    my $timestamp = strftime("[%Y-%m-%d %H:%M:%S]", localtime);
    return "$timestamp $unformatted";
}

sub _isValidFromCli {
    my ($cliOptRef) = @_;
    return (!$ENV{'REMOTE_USER'}
        && ($cliOptRef && $$cliOptRef{'webadc_zconf_upd'} && ('1' eq $$cliOptRef{'webadc_zconf_upd'})));
}

sub validateRoot {
    my ($isJsonOut, $cliOptRef) = @_;
    local $ENV{'REMOTE_USER'} = 'root' if _isValidFromCli($cliOptRef);
    Whostmgr::ACLS::init_acls();
    return 1 if (Whostmgr::ACLS::hasroot());

    if ($isJsonOut) {
        print "Status: 403\r\nContent-type: text/plain\r\n\r\n";
        print Cpanel::JSON::Dump( {
            'success' => 0,
            'msg' => formatMsg("You do not have access to the LiteSpeed ADC ZConf Plugin.")
        } );
    }
    else {
        print "You do not have access to the LiteSpeed ADC ZConf Plugin.\n";
    }
    return;
}

sub sendJson {
    my ($success, $msg) = @_;
    my $cgi = CGI->new();
    my $charset = $cgi->param('charset') || 'utf-8';
    my $json = Cpanel::JSON::Dump(
        {
            'success' => $success,
            'msg'   => formatMsg($msg)
        }
    );

    print $cgi->header('-type' => 'text/plain', '-charset' => $charset,
            '-Content_length' => length(Encode::encode($charset, $json)));
    print $json;
}

sub backupFile {
    my ($fileName, $errorref, $archiveroot, $pattern) = @_;
    my $zip;
    my @backups = glob "${fileName}*";
    my $cnt = @backups; # get array length
    my $backupname = $fileName . '_bak' . $cnt;

    if ($cnt == 0) {
        return 1; # no previous confs existed
    }
    elsif ($cnt < 11) { # current conf and 10 backups
        return (rename($fileName, $backupname));
    }

    my ($fileWithoutSuf, $dir, $suffix) = fileparse($fileName, qr/\.[^.]*/);
    unless ($archiveroot) {
        $archiveroot = $fileWithoutSuf;
    }

    my $archiveName = $dir . $fileWithoutSuf . '_bak.zip';

    $zip = Archive::Zip->new();
    if (($zip->addTreeMatching($dir, $archiveroot, $pattern) != AZ_OK)
      || ($zip->writeToFileNamed($archiveName) != AZ_OK)) {
        $$errorref = "Failed to back up old confs to zip. Will attempt to back up to another file name.";
        return (rename($fileName, $backupname));
    }

    unless (unlink(@backups)) {
        $$errorref = "Failed to delete backed up configurations. Received error: $!";
        return;
    }
    return 1;
}

sub loadPluginConf {
    my ($paramref, $errorref) = @_;

    if (-f $pluginConfFile) {
        my $status = eval {
            $$paramref = retrieve($pluginConfFile);
            1;
        };
        if (!defined($status)) {
            $$errorref = "Retrieve failed, malformed configuration file. $!";
        }
        elsif (!defined($$paramref)) {
            $$errorref = "IO error during configuration file retrieve. Please check the permissions of $pluginConfFile";
        }
        return 1;
    }
=begin comment
#The else case can be used for outputting another message.
    else {
        print "did not exist\n";
    }
=end comment
=cut
    return;
}

sub openLog {
    my $msg;
    my $ret;
    my $rolloverSize = 1024*1024;
    my $size = 0;
    if ($logFH) {
#        zconflib::utils::log("Log file already opened\n");
        return 1;
    }

    if (-f $logFileName) {
        $size = -s $logFileName;
        if ($size > $rolloverSize) {
            return unless (backupFile($logFileName, \$msg,
                                    'logs', 'webadc_zconf\.log'));
        }
    }

    return unless (open($logFH, ">>", $logFileName));

    zconflib::utils::log("Log file rolled over!\n") if ($size > $rolloverSize);

    return 1;
}

sub closeLog {
    return unless $logFH;
    close($logFH);
    undef $logFH;
}

sub log {
    my ($msg) = @_;
    print $logFH formatMsg($msg);
}

1;
